from __future__ import print_function

'''
   setting add-on for mds mp/sp workers.
 
   2014 02 worker type setting is removed. it uses always "mp" worker.

'''

from ifigure.mdsplus.mdsscope import MDSWorkerPool
from ifigure.mdsplus.mds_sp_worker_pool import MDSSPWorkerPool
from ifigure.mdsplus.mds_mpth_worker_pool import MDSMPWorkerPool
from ifigure.utils.setting_parser import iFigureSettingParser as SettingParser
from ifigure.utils.edit_list import DialogEditList
import wx
import subprocess
import wx
import os
import shlex
import socket

module_name = 'MDSplusWorker'
class_name = 'MDSplusWorker'
menu = []  # this will be filled in method below
method = ['onReset', 'onSetting', 'onShowStatus', 'onActivate',
          'onStartWorker', 'init', 'run_job', 'clean',
          'tree_viewer_menu']

icon = 'world_link.png'
can_have_child = False
has_private_owndir = False
allow_only_one_active = True


def init(self, *args, **kargs):
    #    if not kargs.has_key('src'):
    self.td._pool = MDSWorkerPool()
    p = SettingParser()
    p.set_rule('connection', {}, nocheck=True)
    self.td.setvar('setting', p.read_setting('mdsplus.connection_setting'))
#    self.onStartWorker()


def tree_viewer_menu(self):
    if not self.td.is_suppress():
        menu = [
            ("Show Status", "onShowStatus",  False),
            ("Start Worker", "onStartWorker",  False),
            ("Stop Worker",    "onReset",    False),
            ("Setting...", "onSetting",  False)]
    else:
        menu = [("Activate", "onActivate",  False), ]
    return menu


def onActivate(self, evt):
    self.td.activate()
    self.td._pool.reset()
    self.onStartWorker()
    evt.Skip()


def clean(self, *args, **kargs):
    self.td._pool.reset()
    p = SettingParser()
    p.write_setting('mdsplus.connection_setting',
                    self.td.getvar('setting'))


def run_job(self, a):
    from ifigure.mdsplus.mdsscope import print_threaderror
    #self.pool = MDSWorkerPool.pool
    if self.td._pool is None:
        wx.CallAfter(print_threaderror, [
                     'can not run job. worker pool is not ready'])
        return

    maxwait = 25
    wait = 0
    import time
    while(wait < maxwait):
        time.sleep(0.5)
        w, w_id = self.td._pool.get_worker()
        if w is not None:
            a.start_job(self.td._pool, w_id)
            return
        wait = wait+1
    else:
        wx.CallAfter(print_threaderror, [
                     'can not run job. waiting time exceeded max time'])
        return


def onReset(self, e=None):
    self.td._pool.reset()


def onStartWorker(self, e=None):
    s = self.td.getvar('setting')
    c_type = s['connection_type']
    w_type = s['worker']
    if c_type == 'proxy':
        host = s['connection'][0]['server']
        port = s['connection'][0]['port']
        c_type = ('proxy', host, port)
    else:
        c_type = (c_type,)
    t_type = self.td.getvar('translater')
    if t_type is None:
        t_type = 'default'
#    def func(a=c_type, b=w_type):
    self.td._pool.start(c_type, w_type, t_type)


def onShowStatus(self, e=None):
    pool = self.td._pool.pool
    print(pool)
    if isinstance(pool, MDSSPWorkerPool):
        print(pool.pch)
    elif isinstance(pool, MDSMPWorkerPool):
        for w in pool.workers:
            print(w)
    else:  # isinstance(pool, MDSTHWorkerPool):
        for w in pool.workers:
            print(w)


def onSetting(self, e=None):
    import ifigure.mdsplus.mdsscope
    s = self.td.getvar('setting')
    host = s['connection'][0]['server']
    port = s['connection'][0]['port']

    if s['connection_type'] == 'direct':
        txt = 'off'
    else:
        txt = 'on'

    list6 = [["Use Proxy", txt, 1, {"values": ["on", "off"]}],
             ["Root Tree", "cmod", 0, None],
             ["Proxy Server", str(host), 0, None],
             ["Port", str(port), 0, None],
             ["NumWorker", str(
                 ifigure.mdsplus.mdsscope.mds_num_worker), 0, None],
             #           ["Worker", s['worker'], 4, {'style':wx.CB_READONLY, 'choices':['mp', 'th', 'sp']}]]
             ["Worker", s['worker'], 4, {'style': wx.CB_READONLY, 'choices': ['mp', 'th']}]]

    value = DialogEditList(list6, modal=True,
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                           tip=None,
                           parent=self.td.get_app())

    if value[0]:
        v = value[1]
    else:
        return
    if str(v[0]) == 'on':
        t = 'proxy'
    else:
        t = 'direct'
    onum = ifigure.mdsplus.mdsscope.mds_num_worker
    ifigure.mdsplus.mdsscope.mds_num_worker = int(v[4])
    w = str(v[5])
#    w = 'mp' ## always use mp worker
    port = str(v[3])
    host = str(v[2])
    restart = False
    if (s['connection_type'] != t or
        s['worker'] != w or
        onum != ifigure.mdsplus.mdsscope.mds_num_worker or
        s['connection'][0]['server'] != host or
            int(s['connection'][0]['port']) != int(port)):
        restart = True
    s['connection_type'] = t
    s['worker'] = w
    s['connection'][0]['server'] = host
    s['connection'][0]['port'] = int(port)
    if restart:
        self.onReset()
        self.onStartWorker()

    self.td.setvar('setting', s)



from __future__ import print_function
from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.widgets.canvas.file_structure import *
from ifigure.mto.axis_user import XUser, YUser
from ifigure.ifigure_config import pick_r
import ifigure
import os
import ifigure.utils.cbook as cbook
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty, UndoRedoFigobjProperty, UndoRedoFigobjMethod
from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.utils.args_parser import ArgsParser

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigSpline')


def bs_basis(t):
    at = abs(t)
    if t > -1. and t < 1.:
        return (3*at*at*at-6*at*at+4)/6
    if t > -2. and t < 2.:
        return -(at-2)*(at-2)*(at-2)/6
    return 0.


def calc_xy(xp, yp, mesh=10, mode=1):
    if mode == 0:
        return [p for p in xp], [p for p in yp]
    xp1 = np.concatenate(([xp[0], xp[0]], xp, [xp[-1], xp[-1]]))
    yp1 = np.concatenate(([yp[0], yp[0]], yp, [yp[-1], yp[-1]]))

    ks = range(len(xp)+2)
    ts = np.arange((len(xp)+1)*mesh+1)/float(mesh)-1
    xval = np.zeros(len(ts))
    yval = np.zeros(len(ts))
    i = 0
    for t in ts:
        j = 0
        for j in range(len(xp1)):

            t0 = j-2.
#           print t0, t
            xval[i] = xval[i]+xp1[j]*bs_basis(t-t0)
            yval[i] = yval[i]+yp1[j]*bs_basis(t-t0)
            j = j+1
        i = i+1
    return xval, yval


class FigSpline(FigObj, XUser, YUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._objs = []  # for debug....
            obj._hit_idx = -1
            obj._hit_seg = -1
            obj._hit_artist = -1
            obj._mesh = 10
            obj._data_extent = None
            obj._data_extent_checked = False
            obj._sp_interp = 1
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_opt('x', None, ['numbers|nonstr', 'dynamic'])
        p.add_var('y', ['numbers|nonstr', 'dynamic'])

        p.set_ndconvert("x", "y")
        p.set_squeeze_minimum_1D("x", "y")

        p.add_key('draggable', True)

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)

        if v["x"] is None:
            v["x"] = np.arange(v["y"].shape[-1])
        for name in v:
            obj.setvar(name, v[name])
        obj.setvar("kywds", kywds)
        return obj

    def __init__(self, *args,  **kywds):
        XUser.__init__(self)
        YUser.__init__(self)

        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigSpline, self).__init__(*args, **kywds)

    @classmethod
    def isFigSpline(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def get_namebase(self):
        return 'spline'

    @classmethod
    def property_in_file(self):
        return ["color", "linestyle",
                "linewidth", "marker"]

    @classmethod
    def property_in_palette(self):
        return (["spline", "line", "marker"],
                [["spinterp"],
                 ["color", "linestyle", "linewidth"],
                 ["markerfacecolor", "markeredgecolor", "marker"]])

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ["x", "y"]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'plot1.png')
        return [idx1]

    def isDraggable(self):
        return self._var["draggable"]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)

        super(FigSpline, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)

    def args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y"]
        use_np = [True, True]
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)
        x = values[0]
        y = values[1]
        if x is None:
            x = np.arange((y.shape)[-1])
            self.setp("x", x)
        if x is None:
            return False
        if y is None:
            return False
        if (x.ndim == 1 and y.ndim == 1 and
                x.size != y.size):
            self.setp("x", None)
            self.setp("y", None)
            return False
        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        container = self.get_container()

        if self.isempty() is False:
            return
#              self.del_artist(delall=True)

        dprint2("generating spline artist")
        x, y = self._eval_xy()  # this handles "use_var"
        if (x is not None and
                y is not None):

            xval, yval = calc_xy(x, y,
                                 mesh=self._mesh,
                                 mode=self._sp_interp)
            try:
                self._artists = container.plot(xval, yval,
                                               **(self._var["kywds"]))
            except:
                self._artists = []
                raise ValueError("failed to generate spline artist")

        lp = self.getp("loaded_property")
        if lp is not None:
            for i in range(0, len(lp)):
                if i < len(self._artists):
                    self.set_artist_property(self._artists[i], lp[i])
            self.delp("loaded_property")

        for artist in self._artists:
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))
            self._objs.append(weakref.ref(artist))

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                a.remove()
                # container.lines.remove(a)

        super(FigSpline, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        container = self.get_container()
        if val == True:
            for a in alist:
                x = self.getp("x")
                y = self.getp("y")
#              x=a.get_xdata()
#              y=a.get_ydata()
                hl = container.plot(x, y, marker='s',
                                    color='k', linestyle='None',
                                    markerfacecolor='None',
                                    markeredgewidth=0.5,
                                    scalex=False, scaley=False)
                for item in hl:
                    self._objs.append(weakref.ref(item))
                    a.figobj_hl.append(item)
        else:

            for a in alist:
                for hl in a.figobj_hl:
                    hl.remove()
                a.figobj_hl = []
#
#   def hit_test
#

    def picker_a(self, artist, evt):
        self._hit_seg = -1
        self._hit_idx = -1
        self._hit_artist = -1
        axes = artist.axes
        if axes is None:
            return False, {}
        x = self.getp("x")
        y = self.getp("y")
        trans = axes.transData.transform
        itrans = axes.transData.inverted().transform

        # this checks if picker can use xdata/ydata
        # for hit test
        p = np.array([evt.xdata, evt.ydata]).reshape(1, 2)
        pd = trans(p)
        if abs(pd.flat[0] - evt.x) > 3 or abs(pd.flat[1] - evt.y) > 3:
            return False, {}

        for i in range(len(x)):
            x0d, y0d = trans((x[i], y[i]))
            if abs(x0d-evt.x)+abs(y0d-evt.y) < pick_r:
                self._hit_idx = i
                self._hit_artist = (self._artists.index(artist), evt)
                # print "hit control", self._hit_idx
                return True, {'child_artist': artist}

        xdata = artist.get_xdata()
        ydata = artist.get_ydata()
        ans, idx = cpicker.CheckLineHit(xdata, ydata,
                                        evt.xdata, evt.ydata, trans, itrans)

        if ans:
            check = True
            extra = {}

            self._hit_artist = (self._artists.index(artist), evt)
            self._hit_seg = idx
        else:
            check = False
            extra = {}
        if check:
            return True, {'child_artist': artist}
        return False, {}

    def picker_a0(self, artist, evt):
        hit, extra = self.picker_a(artist, evt)
        if hit:
            return hit, extra, 'area', 3
        else:
            return False, {}, None, 0
#
#   drag
#

    def dragstart_a(self, a, evt):
        return self.dragstart(a, evt)

    def dragstart(self, a, evt):
        #        print "drag spline", self._hit_idx
        self._drag_start = evt

    def get_dragged_node(self, evt):
        d = (evt.xdata - self._drag_start.xdata,
             evt.ydata - self._drag_start.ydata)

        if self._hit_idx != -1:  # change control
            x = [item for item in self.getp("x")]
            y = [item for item in self.getp("y")]
            x[self._hit_idx] = x[self._hit_idx] + d[0]
            y[self._hit_idx] = y[self._hit_idx] + d[1]
        else:  # transpose
            x = [item + d[0] for item in self.getp("x")]
            y = [item + d[1] for item in self.getp("y")]
        return x, y

    def drag_a(self, a, evt, shift=None, scale=None):
        return self.drag(a, evt), scale

    def drag(self, a, evt):
        if evt.inaxes is None:
            return 0
        x, y = self.get_dragged_node(evt)
        if len(a.figobj_hl) == 1:
            if a.figobj_hl[0] is not None:
                a.figobj_hl[0].set_xdata(x)
                a.figobj_hl[0].set_ydata(y)
        xval, yval = calc_xy(x, y, mesh=self._mesh,
                             mode=self._sp_interp)
        a.set_xdata(xval)
        a.set_ydata(yval)
        return 1

    def drag_a_get_hl(self, a):
        return self.drag_get_hl(a)

    def drag_get_hl(self, a):
        self._alpha_backup = a.get_alpha()
        a.set_alpha(0.5)
        return [a]

    def drag_a_rm_hl(self, a):
        self.drag_rm_hl(a)

    def drag_rm_hl(self, a):
        a.set_alpha(self._alpha_backup)

    def dragdone_a(self, a, evt, shift=None, scale=None):
        return self.dragdone(a, evt), scale

    def dragdone(self, a, evt):
        if evt.inaxes is None:
            return 0

        x, y = self.get_dragged_node(evt)
#        app = evt.guiEvent.GetEventObject().GetTopLevelParent()
#        hist = app.history
        window = evt.guiEvent_memory.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)
        hist.start_record()
        action1 = UndoRedoFigobjMethod(a, 'splinenode', (x, y))
        hist.add_history(action1)
        hist.stop_record()
        return 1
#
#  Popup in Canvas
#

    def canvas_menu(self):
        menus = [("Add Node",  self.onAddNode, None),
                 ("Remove Node", self.onRmNode, None),
                 ("Export Node", self.onExportNode, None)] + super(FigSpline, self).canvas_menu()
        return menus

    def onAddNode(self, evt):
        if self._hit_artist == -1:
            return
        if self._hit_seg == -1:
            return

        # _hit_artist = (artist index, mpl event)
        idx = int(self._hit_seg/self._mesh)
        x = [p for p in self.getp("x")]
        y = [p for p in self.getp("y")]
        x.insert(idx, self._hit_artist[1].xdata)
        y.insert(idx, self._hit_artist[1].ydata)

        window = evt.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)
        hist.start_record()
        a = self._artists[self._hit_artist[0]]
        action1 = UndoRedoFigobjMethod(a,
                                       'splinenode', (x, y))
        hist.add_history(action1)
        hist.stop_record()

    def onRmNode(self, evt):
        if self._hit_artist == -1:
            return
        if self._hit_idx == -1:
            return

        x = self.getp("x")
        y = self.getp("y")
        xn = [x[i] for i in range(len(x)) if i != self._hit_idx]
        yn = [y[i] for i in range(len(y)) if i != self._hit_idx]

#        app = evt.GetEventObject().GetTopLevelParent()
        window = evt.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)
        hist.start_record()
        a = self._artists[self._hit_artist[0]]
        action1 = UndoRedoFigobjMethod(a,
                                       'splinenode', (xn, yn))
        hist.add_history(action1)
        hist.stop_record()

    def onExportNode(self, event):
        fig_val = {"xdata": self.getp("x"),
                   "ydata": self.getp("y")}
        text = '#Exporting data as fig_val[\'xdata\'], fig_val[\'ydata\']\"'
        self._export_shell(fig_val, 'fig_val', text)

    def onExport(self, event):
        from matplotlib.artist import getp
        fig_val = None
        canvas = event.GetEventObject()
        sel = [a() for a in canvas.selection]
        for a in self._artists:
            if a in sel:
                print("Exporting Data to Shell")
                fig_val = {"xdata": getp(a, "xdata"),
                           "ydata": getp(a, "ydata")}
                break
        if fig_val is not None:
            text = '#Exporting data as fig_val[\'xdata\'], fig_val[\'ydata\']\"'
            self._export_shell(fig_val, 'fig_val', text)

    def set_splinenode(self, value, a):
        self.setp("x", value[0])
        self.setp("y", value[1])
        xval, yval = calc_xy(value[0], value[1],
                             mesh=self._mesh,
                             mode=self._sp_interp)
        a.set_xdata(xval)
        a.set_ydata(yval)
        self._data_extent = None
        self._data_extent_checked = False

    def get_splinenode(self, a):
        x = self.getp("x")
        y = self.getp("y")
        return x, y
#
#   data extent
#

    def get_data_extent(self):
        if (self._data_extent is not None and
                self._data_extent_checked):
            return self._data_extent
        if (self._data_extent is not None and
                self.isempty()):
            return self._data_extent
        x, y = self._eval_xy()
        if self.isempty():
            if x is None:
                self._data_extent = [0, len(y), np.min(y), np.max(y)]
            else:
                self._data_extent = [
                    np.min(x), np.max(x), np.min(y), np.max(y)]
        else:
            xr = (np.inf, -np.inf)
            yr = (np.inf, -np.inf)
            for a in self._artists:
                xdata = a.get_xdata()
                ydata = a.get_ydata()
                xr = (min((xr[0], np.amin(xdata))),
                      max((xr[1], np.amax(xdata))))
                yr = (min((yr[0], np.amin(ydata))),
                      max((yr[1], np.amax(ydata))))

            self._data_extent = [min((min(xr), np.amin(x))),
                                 max((max(xr), np.amax(x))),
                                 min((min(yr), np.amin(y))),
                                 max((max(yr), np.amax(y)))]
            self._data_extent_checked = True
        return self._data_extent
#
#   range
#

    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y = self._eval_xy()
        if len(self._artists) != 0:
            xc = self._artists[0].get_xdata()
        else:
            xc, yc = calc_xy(x, y,
                             mesh=self._mesh,
                             mode=self._sp_interp)
        if scale == 'log':
            x = mask_negative(x)
            xc = mask_negative(xc)
        return self._update_range(xrange, [np.nanmin([np.nanmin(x), np.nanmin(xc)]),
                                           np.nanmax([np.nanmax(x), np.nanmax(xc)])])

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        x, y = self._eval_xy()
        if len(self._artists) != 0:
            xc = self._artists[0].get_xdata()
            yc = self._artists[0].get_ydata()
        else:
            xc, yc = calc_xy(x, y,
                             mesh=self._mesh,
                             mode=self._sp_interp)
        ym = np.ma.masked_array(y)
        ycm = np.ma.masked_array(yc)
        ym[x < xrange[0]] = np.ma.masked
        ym[x > xrange[1]] = np.ma.masked
        ycm[xc < xrange[0]] = np.ma.masked
        ycm[xc > xrange[1]] = np.ma.masked
        if scale == 'log':
            ym[x < 0] = np.ma.masked
            ycm[xc < 0] = np.ma.masked
        return self._update_range(yrange, [np.nanmin([np.nanmin(ym), np.nanmin(ycm)]),
                                           np.nanmax([np.nanmax(ym), np.nanmax(ycm)])])

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        var = {'x': check(self, 'x'),
               'y': check(self, 'y')}

        if var["x"] is not True:
            var["xdata"] = self.getp("x")
        if var["y"] is not True:
            var["ydata"] = self.getp("y")
        var["_sp_interp"] = self._sp_interp

        data['FigSpline'] = (1, var)
        data = super(FigSpline, self).save_data2(data)
        return data

    def load_data2(self, data):
        d = data['FigSpline']
        super(FigSpline, self).load_data2(data)
        var = d[1]
        names = ["x", "y"]
        if '_sp_interp' in var:
            self._sp_interp = var['_sp_interp']
        for name in names:
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])

    def _eval_xy(self):
        names = ("x", "y")
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return [None]*len(names)
        return self.getp(names)

    def set_sp_interp(self, value, a):
        print(value)
        if str(value) == 'linear':
            self._sp_interp = 0
        else:
            self._sp_interp = 1

        xval, yval = calc_xy(self.getp('x'),
                             self.getp('y'),
                             mesh=self._mesh,
                             mode=self._sp_interp)
        a.set_xdata(xval)
        a.set_ydata(yval)
        self._data_extent = None
        self._data_extent_checked = False

    def refresh_artist_data(self):
        super(FigSpline, self).refresh_artist_data()
        self._data_extent_checked = False

    def get_sp_interp(self, a):
        if self._sp_interp == 0:
            return 'linear'
        else:
            return 'spline'

    def setp(self, *args):  # arg = name, var or var
        super(FigSpline, self).setp(*args)
        if len(args) == 2:
            if (args[0] == 'x' or args[0] == 'y'):
                self.setvar(args[0], args[1])

# Public
    def GetSplineData(self, x=None):
        from matplotlib.artist import getp
        return [(getp(a, "xdata"),
                 getp(a, "ydata")) for a in self._artists]



from __future__ import print_function
from six.moves.queue import Queue, Empty
import wx.py.shell  # (wx4 removed this) wx.lib.shell
from threading import Timer, Thread
import time
from ifigure.utils.wx3to4 import isWX3
from os.path import expanduser, abspath
import wx
import ifigure
import os
import sys
import ifigure.utils.pickle_wrapper as pickle
import ifigure.server
import numpy as np

import ifigure.utils.debug as debug
from ifigure.mto.treedict import TreeDict
from ifigure.utils.cbook import text_repr
# shell_variable
EvtShellEnter = wx.NewEventType()
EVT_SHELL_ENTER = wx.PyEventBinder(EvtShellEnter, 1)

try:
    from ifigure.version import ifig_version
except:
    ifig_version = 'dev'

dprint1, dprint2, dprint3 = debug.init_dprints('SimpleShell')


ON_POSIX = 'posix' in sys.builtin_module_names

sx_print_to_consol = False


def enqueue_output(p, queue):
    while True:
        line = p.stdout.readline()
        queue.put(line)
        if p.poll() is not None:
            queue.put(p.stdout.read())
            break
    queue.put('process terminated')


def run_in_thread(p):
    q = Queue()
    t = Thread(target=enqueue_output, args=(p, q))
    t.daemon = True  # thread dies with the program
    t.start()

    lines = ["\n"]
    line = ''
    alive = True
    app = wx.GetApp().TopWindow
    if sx_print_to_consol:
        write_cmd = app.proj_tree_viewer.consol.log.AppendText
    else:
        write_cmd = app.shell.WriteTextAndPrompt
    while True:
        time.sleep(0.01)
        try:
            line = q.get_nowait()  # or q.get(timeout=.1)
        except Empty:
            if len(lines) != 0:
                wx.CallAfter(write_cmd, ''.join(lines))
            lines = []
        except:
            import traceback
            traceback.print_exc()
            break
        else:  # got line
            lines.append(line)
        if line.startswith('process terminated'):
            if len(lines) > 1:
                wx.CallAfter(write_cmd, ''.join(lines[:-1]))
            break
    return


def sx(strin=''):
    if strin == '':
        return
    if strin.startswith('cd '):
        dest = strin[3:]
        dest = abspath(expanduser(dest))
        os.chdir(dest)
        txt = os.getcwd()
    else:
        import subprocess as sp

        p = sp.Popen(strin, shell=True, stdout=sp.PIPE,
                     stderr=sp.STDOUT, universal_newlines=True)
        t = Thread(target=run_in_thread, args=(p,))
        t.daemon = True  # thread dies with the program
        t.start()
        # run_in_thread(p)
        #txt = ''.join(p.stdout.readlines())


class ShellBase(wx.py.shell.Shell):
    def Paste(self):

        if sys.platform == 'darwin':
            import wx
            data = wx.TextDataObject()
            if wx.TheClipboard.Open():
                wx.TheClipboard.GetData(data)
                txt = data.GetText()
                wx.TheClipboard.Close()
            else:
                txt = ''
            self.write(txt)
        else:
            return super(ShellBase, self).Paste()

    def OnKeyDown(self, evt):

        if evt.GetKeyCode() == wx.WXK_UP:
            self.OnHistoryReplace(1)
            return
        if evt.GetKeyCode() == wx.WXK_DOWN:
            self.OnHistoryReplace(-1)
            return

        if hasattr(evt, 'RawControlDown'):
            mod = (evt.RawControlDown() +
                   evt.AltDown()*2 +
                   evt.ShiftDown()*4 +
                   evt.MetaDown()*8)
        else:
            mod = (evt.ControlDown() +
                   evt.AltDown()*2 +
                   evt.ShiftDown()*4 +
                   evt.MetaDown()*8)

        if evt.GetKeyCode() == 78 and mod == 1:  # control + N (next command)
            #          self.OnHistoryReplace(-1)
            self.LineDown()
            return
        if evt.GetKeyCode() == 80 and mod == 1:  # ctrl + P (prev command)
            #          self.OnHistoryReplace(1)
            self.LineUp()
            return
        if evt.GetKeyCode() == 68 and mod == 1:  # ctrl + D (delete)
            self.CharRight()
            self.DeleteBackNotLine()
            self.st_flag = True
            return
        if evt.GetKeyCode() == 70 and mod == 1:  # ctrl + F (forward)
            self.CharRight()
            return
        if evt.GetKeyCode() == 66 and mod == 1:  # ctrol + B (back)
            self.CharLeft()
            return
        if evt.GetKeyCode() == 89 and mod == 1:  # ctrl + Y (paste)
            self.Paste()
            self.st_flag = True
            return
        if evt.GetKeyCode() == 87 and mod == 1:  # ctrl + W (cut)
            self.Cut()
            self.st_flag = True
            return
        if evt.GetKeyCode() == 87 and mod == 2:  # alt + W (copy)
            self.Copy()
            self.st_flag = True
            return
        if evt.GetKeyCode() == 90 and mod == 1:  # ctrl + Z (cut)
            self.st_flag = True
            import ifigure.events
            ifigure.events.SendUndoEvent(self.lvar["proj"], w=self)
            return
        if evt.GetKeyCode() == 69 and mod == 1:  # ctrl + E (end of line)
            self.LineEnd()
            return
        if evt.GetKeyCode() == 75 and mod == 1:  # ctrl + K (cut the rest)
            self.LineEndExtend()
            self.Cut()
            self.st_flag = True
            return
        if evt.GetKeyCode() == 65 and mod == 1:  # ctrl + A (beginning)
            self.Home()
            self.CharRight()
            self.CharRight()
            self.CharRight()
            self.CharRight()
            return

        if evt.GetKeyCode() == wx.WXK_RETURN:
            self.st_flag = True

        if (evt.ControlDown() == False and
            evt.AltDown() == False and
            evt.MetaDown() == False and
            evt.GetKeyCode() > 31 and
                evt.GetKeyCode() < 127):
            self.st_flag = True
#       print 'setting search text', self.st, self.historyIndex
        super(ShellBase, self).OnKeyDown(evt)


class simple_shell_droptarget(wx.TextDropTarget):
    def __init__(self, obj):
        wx.TextDropTarget.__init__(self)
        self.obj = obj

    def OnDropText(self, x, y, indata):
        app = self.obj.GetTopLevelParent()
        txt = app._text_clip
        app._text_clip = ''
#        print txt, x, y
        txt.strip('\n\r')
        txt = '\n... '.join(txt.split('\n'))

        pos = self.obj.PositionFromPoint(wx.Point(x, y))
        self.obj.GotoPos(pos)
        self.obj.write(txt)
        self.obj.GotoPos(pos+len(txt))
        if isWX3:
            wx.CallAfter(self.obj.SetSTCFocus, True)
            wx.CallLater(100, self.obj.SetFocus)
        return True
        # return super(simple_shell_droptarget, self).OnDropText(x, y, indata)

    def OnDragOver(self, x, y, default):
        self.obj.DoDragOver(x, y, default)
        return super(simple_shell_droptarget, self).OnDragOver(x, y, default)


class FakeSimpleShell(wx.Panel):
    def __init__(self, parent=None, *args, **kargs):
        super(FakeSimpleShell, self).__init__(parent, *args, **kargs)
        self.ch = None  # command history panel

#       self.chistory=collections.deque(maxlen=100)
        self.lvar = {}
        self._no_record = False

    def set_command_history(self, panel):
        self.ch = panel

    def set_proj(self, proj):
        self.lvar["proj"] = proj

    def set_tipw(self, w):
        self.tipw = w

    def write_history(self):
        # save last 300 command history
        from ifigure.ifigure_config import rcdir
        file = os.path.join(rcdir, "command_history")
        f = open(file, 'wb')
        f.close()

from wx.py.interpreter import Interpreter
class MyInterp(Interpreter):
    '''
    runsource is modified to perfrom the runsource with either symbol='exec' or 'single'

    'single' is an orignal mode
    'exec' is used to run multiple statement at once. this is useful when we run
     segment of script written in scripteditor
    '''
    def __init__(self, *args, **kwargs):
        Interpreter.__init__(self, *args, **kwargs)
        self._runsouce_mode = 'single'

    def set_single_run_mode(self):
        self._runsouce_mode = 'single'

    def set_batch_run_mode(self):
        self._runsouce_mode = 'exec'

    def runsource(self, source):
        """Compile and run source code in the interpreter."""
        stdin, stdout, stderr = sys.stdin, sys.stdout, sys.stderr
        sys.stdin, sys.stdout, sys.stderr = \
                   self.stdin, self.stdout, self.stderr

        # we do this using "exec"
        more = super(Interpreter, self).runsource(source, symbol=self._runsouce_mode)

        # this was original
        #more = InteractiveInterpreter.runsource(self, source)

        # If sys.std* is still what we set it to, then restore it.
        # But, if the executed source changed sys.std*, assume it was
        # meant to be changed and leave it. Power to the people.
        if sys.stdin == self.stdin:
            sys.stdin = stdin
        else:
            self.stdin = sys.stdin
        if sys.stdout == self.stdout:
            sys.stdout = stdout
        else:
            self.stdout = sys.stdout
        if sys.stderr == self.stderr:
            sys.stderr = stderr
        else:
            self.stderr = sys.stderr
        return more

class SimpleShell(ShellBase):
    #   arrow key up/down are trapped to work as history
    #   this might be possible by configuration file!?
    #
    SHELL = None

    def __init__(self, parent=None):
        self.ch = None  # command history panel

#       self.chistory=collections.deque(maxlen=100)
        self.lvar = {}
        self._no_record = False

        sc = os.path.join(os.path.dirname(ifigure.__file__), 'startup.py')
        txt = '    --- Welcome to piScope ('+ifig_version+')---'
        super(SimpleShell, self).__init__(parent,
                                          locals=self.lvar,
                                          startupScript=sc,
                                          introText=txt,
                                          InterpClass=MyInterp)

        if os.getenv('PYTHONSTARTUP') is not None:
            file = os.getenv('PYTHONSTARTUP')
            if os.path.exists(file):
                dprint1('running startup file', file)
                txt = 'Running user startup file '+file
                self.push('print %r' % txt)
                #self.execfile(file, globals(), self.lvar)
                self.execStartupScript(file)

        self.SetDropTarget(simple_shell_droptarget(self))

        from ifigure.ifigure_config import rcdir
        file = os.path.join(rcdir, "command_history")
        try:
            f = open(file, 'rb')
            self.history = pickle.load(f)
            f.close()
        except Exception:
            import traceback
            traceback.print_exc()
            print("Can not load command history file")

        if self.history[-1] != '#end of history':
            self.history.append('#end of history')

        SimpleShell.SHELL = self
        self.Bind(wx.EVT_LEFT_DOWN, self.onLeftDown)
        self.Bind(wx.EVT_LEFT_UP, self.onLeftUp)
        self.Bind(wx.EVT_SET_FOCUS, self.onSetFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.onKillFocus)
        self.st = ''  # search txt
        self.st_flag = False

        self._auto_complete = True

    def onSetFocus(self, evt):
        evt.Skip()

    def onKillFocus(self, evt):
        evt.Skip()

    def setBuiltinKeywords(self):
        '''
        this overwrite the origial setBuiltinKeywords
        '''
        from six.moves import builtins
        builtins.exit = builtins.quit = \
            self.quit
        builtins.forceexit = builtins.forcequit = \
            self.forcequit

        builtins.sx = sx

    def set_command_history(self, panel):
        self.ch = panel

    def quit(self):
        '''
        exit piscope
        '''
        self.GetTopLevelParent().onQuit()

    def forcequit(self):
        '''
        Terminate piScope (kill command is called)
        '''
        import os
        pid = os.getpid()
        os.system('kill '+str(pid))

    def set_proj(self, proj):
        self.lvar["proj"] = proj

    def write_history(self):
        # save last 300 command history
        from ifigure.ifigure_config import rcdir
        file = os.path.join(rcdir, "command_history")
        f = open(file, 'wb')
        h = self.history[0:1000]
        h.append(self.history[-1])
        pickle.dump(h, f)
        f.flush()
        f.close()

    def autoCompleteShow(self, command, offset=0):
        #
        #   add treeobject method to menu
        #   only does autocomplete when caret is at the end of line
        #
        if self._auto_complete:
            if (self.GetCurrentPos() !=
                    self.GetLineEndPosition(self.GetCurrentLine())):
                return
            """Display auto-completion popup list."""
            self.AutoCompSetAutoHide(self.autoCompleteAutoHide)
            self.AutoCompSetIgnoreCase(self.autoCompleteCaseInsensitive)
            list = self.interp.getAutoCompleteList(command,
                                                   includeMagic=self.autoCompleteIncludeMagic,
                                                   includeSingle=self.autoCompleteIncludeSingle,
                                                   includeDouble=self.autoCompleteIncludeDouble)
            try:
                self.lvar['_tmp_'] = None
                txt = 'if isinstance(command, TreeDict): _tmp_=' + \
                    command+'get_children()'
                code = compile(txt, '<string>', 'exec')
                exec(code, globals(), self.lvar)
            except:
                pass

            if self.lvar['_tmp_'] is not None:
                list = list + [x[0] for x in self.lvar['_tmp_']]
                list.sort(lambda x, y: cmp(x.upper(), y.upper()))
            if list:
                options = ' '.join(list)
                #offset = 0
#                 self.AutoCompSetMaxWidth(3)
                self.AutoCompShow(offset, options)

#            super(SimpleShell, self).autoCompleteShow(command, offset=offset)

    def autoCallTipShow(self, command, insertcalltip=True, forceCallTip=False):
        #        super(SimpleShell, self).autoCallTipShow(*args, **kargs)
        #        return
        (name, argspec, tip) = self.interp.getCallTip(command)
        self._no_record = False
        self.tipw.update(name, argspec, tip)
        if not self.autoCallTip and not forceCallTip:
            return
        if not self._auto_complete:
            return
        startpos = self.GetCurrentPos()
        if argspec and insertcalltip and self.callTipInsert:
            self.write(argspec + ')')
            endpos = self.GetCurrentPos()
            self.SetSelection(startpos, endpos)
        return

        name = str(command).split('(')[-2]
        s = ''
        for c in name:
            if not c.isalnum() and c != '.' and c != '_':
                s = c
        if s != '':
            name = name.split(s)[-1]

        self._no_record = True
        self.lvar['_tmp_'] = None
        txt = '_tmp_='+name
        try:
            code = compile(txt, '<string>', 'exec')
            exec(code, globals(), self.lvar)
        except:
            #           dprint1(txt)
            pass
#        try:
#           self.push(txt, True)
#        except:
#           pass
        self._no_record = False
        if '_tmp_' in self.lvar:
            self.tipw.update(self.lvar['_tmp_'])

    def set_tipw(self, w):
        self.tipw = w

    def OnHistoryInsert(self, step):
        #       print "history insert", step
        super(SimpleShell, self).OnHistoryInsert(step)

    def OnHistoryReplace(self, step):
        # print "history replace", step
        if self.st_flag:
            txt, pos = self.GetCurLine()
            self.st = self.getCommand(txt)
            self.st_flag = False
#       print self.st, self.historyIndex
        if self.st != '':
            tmp_idx = self.historyIndex
            rep = 0
            fail_flag = False
            while (-2 < tmp_idx < len(self.history)):
                rep = rep + step
                tmp_idx = tmp_idx + step
                try:
                    st = self.history[tmp_idx]
                    if st[0:len(self.st)] == self.st:
                        break
                except:
                    fail_flag = True
            if not fail_flag:
                super(SimpleShell, self).OnHistoryReplace(rep)
            return
#          while (-2 < self.historyIndex < len(self.history)):
#             super(SimpleShell, self).OnHistoryReplace(step)
#             if self.historyIndex == -1: break
#             if self.historyIndex == len(self.history)-1: break
#             try:
#                 st = self.history[self.historyIndex]
#             except:
#                 print len(self.history), self.historyIndex
#                 import traceback
#                 traceback.print_exc()
#                 break
#             if st[0:len(self.st)] == self.st:
#                 break
        else:
            super(SimpleShell, self).OnHistoryReplace(step)

    def list_func(self):
        llist = self.list_locals()
        for objname, t in llist:
            if t == 'function':
                print(objname)

    def get_shellvar(self, name):
        return self.lvar[name]

    def set_shellvar(self, name, value):
        self.lvar[name] = value

    def del_shellvar(self, name):
        del self.lvar[name]

    def has_shellvar(self, name):
        return name in self.lvar

    def list_locals(self):
        import types

        llist = []
        # tlist = {getattr(types, name):
        #         str(getattr(types, name)).split("'")[1]
        #         for name in dir(types) if not name.startswith('_')}

        for key in self.lvar:
            if not key.startswith('_'):
                val = self.lvar[key]
                t0 = type(val).__name__
                text = text_repr(val)
                try:
                    hasshape = hasattr(val, 'shape')
                except:
                    import traceback
                    traceback.print_exc()
                    hasshape = False

                if hasshape:
                    llist.append((key, t0, text, str(val.shape)))
                else:
                    llist.append((key, t0, text, ''))

                #                for t, name in tlist:
                #                    if t0 == t:
                #                       llist.append((key, name,))
                #                       break
        return llist

    def list_module(self):
        pass

    def execute_text(self, text):
        self.Execute(text)
        return

    def execute_and_hide_main(self, text):
        self.Execute(text)
        self.Execute('import wx;wx.GetApp().TopWindow.goto_no_mainwindow()')

    def onLeftDown(self, e):
        self.Bind(wx.EVT_MOTION, self.onDragInit)
        e.Skip()

    def onLeftUp(self, e):
        self.Unbind(wx.EVT_MOTION)
        e.Skip()

    def onDragInit(self, e):
        self.Unbind(wx.EVT_MOTION)
        sel = self.GetSelectedText()
        if sel == '':
            e.Skip()
            return
        """ Begin a Drag Operation """
        # Create a Text Data Object, which holds the text that is to be dragged
        # app=wx.GetApp()
        p = self
        while p.GetParent() is not None:
            p = p.GetParent()

        p._text_clip = sel

        tdo = wx.TextDataObject(sel)
        tds = wx.DropSource(self)
        tds.SetData(tdo)
        tds.DoDragDrop(True)

    def SendShellEnterEvent(self):
        evt = wx.PyCommandEvent(EvtShellEnter)
        handler = self.GetEventHandler()
        wx.PostEvent(handler, evt)

    def addHistory(self, *args, **kargs):
        #        print  args, kargs
        if args[0].startswith('sx'):
            args = list(args)
            try:
                args[0] = '!'+args[0].split('"')[1]
            except IndexError:
                return
            args = tuple(args)
        if self._no_record:
            return
        wx.py.shell.Shell.addHistory(self, *args, **kargs)
#        print wx.GetApp().IsMainLoopRunning()
        if wx.GetApp().IsMainLoopRunning():
            self.SendShellEnterEvent()
        if self.ch is not None:
            try:
                self.ch.append_text(args[0])
            except UnicodeError:
                print("unicode error")
                pass

    def toggle_overtype(self, evt):
        value = not self.GetOvertype()
        self.SetOvertype(value)

    def GetContextMenu(self):
        menu = super(SimpleShell, self).GetContextMenu()
        menu.AppendSeparator()
        if self.tipw.IsShown():
            f1 = menu.Append(wx.ID_ANY, "Hide Help")
            self.Bind(wx.EVT_MENU, self.onHideHelp, f1)
        else:
            f1 = menu.Append(wx.ID_ANY, "Show Help")
            self.Bind(wx.EVT_MENU, self.onShowHelp, f1)

        if self.GetOvertype():
            f1 = menu.Append(wx.ID_ANY, "Insert Mode")
            self.Bind(wx.EVT_MENU, self.toggle_overtype, f1)
        else:
            f1 = menu.Append(wx.ID_ANY, "Overwrite Mode")
            self.Bind(wx.EVT_MENU, self.toggle_overtype, f1)

        if self._auto_complete:
            f2 = menu.Append(wx.ID_ANY, "Auto Complete Off")
            self.Bind(wx.EVT_MENU, self.onAutoCompOff, f2)
        else:
            f2 = menu.Append(wx.ID_ANY, "Auto Complete On")
            self.Bind(wx.EVT_MENU, self.onAutoCompOn, f2)

        return menu

    def onShowHelp(self, evt):
        self.tipw.Show()

    def onHideHelp(self, evt):
        self.tipw.Hide()

    def onAutoCompOn(self, evt):
        self._auto_complete = True

    def onAutoCompOff(self, evt):
        self._auto_complete = False

    def WriteTextAndPrompt(self, txt):
        self.WriteText(txt)
        self.prompt()



'''
 addon_utils:
       utility functions commonly used in addons

       onOpenOrg : open original file
       onOpenCurrent : open current file
       onLoadFile : ask user an input and copy a file to
                    owndir as working file.
                    it also ask if copy original file into
                    owndir as original file. this is useful
                    if work is done at different machines.

'''
import os
import wx
import shutil
import tempfile
import ifigure.events
from ifigure.utils.edit_list import DialogEditList


def onOpenOrg(td):
    import ifigure.events

    file = td.path2fullpath('ofile_pathmode',
                            'ofile_path')
    if file == '':
        return
    ifigure.events.SendEditFileEvent(td, None, file, readonly=True)


def onWriteFile(td, filename='', dir='', txt='',
                message='Select File to Save',
                wildcard='Any|*'):
    if filename == '':
        open_dlg = wx.FileDialog(None,
                                 message=message,
                                 defaultDir=dir,
                                 style=wx.FD_SAVE,
                                 wildcard=wildcard)
        if open_dlg.ShowModal() != wx.ID_OK:
            open_dlg.Destroy()
            return
        else:
            filename = str(open_dlg.GetPath())
            open_dlg.Destroy()

    with tempfile.NamedTemporaryFile(
            'w', dir=os.path.dirname(filename), delete=False) as tf:
        tf.write(txt)
        tempname = tf.name
    if not os.path.isabs(filename):
        filename = os.path.join(td.owndir(), filename)
    if os.path.exists(filename):
        os.remove(filename)
    shutil.move(tempname, filename)
#    fid = open(filename, 'w')
#    print "writing contents to ", filename
#    fid.write(txt)
#    fid.close()


def onOpenCurrent(td, modename='addon_pathmode',
                  pathname='addon_path'):

    import ifigure.events
    file = td.path2fullpath(modename, pathname)
    ifigure.events.SendEditFileEvent(td, None, file, readonly=False)


def onLoadFile(td, message="Select File",
               modename='addon_pathmode',
               pathname='addon_path',
               extname='addon_ext',
               wildcard='Any|*',
               ask_org_copy=True,
               file=None,
               reject_loc=None):
    '''
    onLoadFile ask user an file to read.
    copy it to its own directory as filename.
    also another copy is made as filename + '.org'
    for future record

    if ask_org_copy = False: it automatically copy original for
    read only data and make a duplicate file for edit. It is only
    useful when the file size is small
    '''
    if td.getvar("original file") is not None:
        dir = os.path.dirname(td.getvar("original file"))
    else:
        dir = '.'
    if not os.path.exists(dir):
        dir = ''

    if file is None:
        open_dlg = wx.FileDialog(None,
                                 message=message,
                                 defaultDir=dir,
                                 style=wx.FD_OPEN, wildcard=wildcard)
        if open_dlg.ShowModal() != wx.ID_OK:
            open_dlg.Destroy()
            return False
        file = open_dlg.GetPath()
        open_dlg.Destroy()

    if reject_loc is not None:
        rmode, rpath = td.fullpath2path(td, file)
        if rmode in reject_loc:
            m = 'Improper import source location'
            ret = dialog.message(None, message=m,
                                 title='Import error')
            return

    try:
        # this may fail if owndir does not exist
        #print(os.path.dirname(file), td.owndir())
        samefile = os.path.samefile(os.path.dirname(file), td.owndir())
        # print(samefile)
    except BaseException:
        samefile = False
    if ask_org_copy and not samefile:
        from ifigure.widgets.dlg_fileimportmode import DlgFileimportmode
        copy_file, pathmodes, ret = DlgFileimportmode(td,
                                                      ask_copyorg=True)

        '''
        choices = ["auto", "abs", "home", "proj"]
        if td.get_extfolderpath() is not None:
            choices.append(td.get_extfolderpath())
        list6 = [[None,  [True, ['auto']], 127,
                 [{'text':'copy file to project'},
                  {'elp':[['Select path mode', 'auto', 4,
                           {"style":wx.CB_READONLY,
                            "choices": choices}],]}], ],
                 [None, False, 3,
                 {"text":"copy original to project as a separate file"}], ]
        value = DialogEditList(list6, modal = True,
                     style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER,
                     tip = None,
                     parent=None,)

        if not value[0]: return
        ret = value[1][1]
        copy_file = value[1][0][0]
        pathmodes = value[1][0][1]
        if str(pathmodes[0]) == 'auto':
           pathmodes = ['proj', 'home', 'abs']
        '''
    else:
        ret = True
        copy_file = True

    for name, child in td.get_children():
        child.destroy()
    if not td.has_owndir():
        td.mk_owndir()
    od = td.owndir()

    if ret:
        new_ofile = os.path.join(od, os.path.basename(file) + '.org')
        shutil.copyfile(file, new_ofile)
        td.setvar('ofile_pathmode', 'owndir')
        td.setvar('ofile_path', os.path.basename(new_ofile))
        if not 'ofile_path' in td._items:
            td._items.append('ofile_path')
#         td._items.append('ofile_path')
    else:
        new_ofile = file
        td.setvar('ofile_pathmode', 'abs')
        td.setvar('ofile_path', file)
        if 'ofile_path' in td._items:
            td.remove_ownitem(items=['ofile_path'])
            td._items.remove('ofile_path')

    # nl_file is the file to be edited
    if copy_file:
        nl_file = os.path.join(od, os.path.basename(file))
        print(nl_file)
        try:
            # this may fail if owndir does not exist
            if os.path.exists(nl_file):
                samefile = os.path.samefile(file, nl_file)
            else:
                samefile = True
        except BaseException:
            import traceback
            traceback.print_exc()
            samefile = False
        if not samefile:
            print("not the same file")
            #td.remove_ownitem(items=[pathname])
            sss = nl_file.split('.')
            nl_file = '.'.join(sss[:-1])+'1.'+sss[-1]
        shutil.copyfile(file, nl_file)
        td.set_path_pathmode(nl_file, modename, pathname, extname)
    else:
        td.remove_ownitem(items=[pathname])
        td.set_path_pathmode(file, modename, pathname,
                             extname, checklist=pathmodes)
    return True



from setuptools import setup
import os


def long_description():
    rootdir = os.path.abspath(os.path.dirname(__file__))    
    with open(os.path.join(rootdir, 'README.md'), encoding='utf-8') as f:
        return f.read()

def run_setup():
    setup(
       long_description = long_description(),
       long_description_content_type = "text/markdown",)

def main():
    run_setup()


if __name__ == '__main__':
    main()



import numpy as np
from matplotlib import cm
from ifigure.mto.fig_surface import FigSurface
from ifigure.utils.triangulation_wrapper import tri_args
from matplotlib.cm import ScalarMappable
from matplotlib.colors import ColorConverter
cc = ColorConverter()


class FigTrisurface(FigSurface):
    def __new__(self, *argc, **kywds):
        if len(argc) == 2 or len(argc) == 4:
            tri = argc[0]
            argc = argc[1:]
        else:
            tri = None
        obj = FigSurface.__new__(self, *argc, **kywds)

        if not hasattr(obj, '_tri'):
            obj._tri = None

        return obj

    def __init__(self, *argc, **kywds):
        if len(argc) == 2 or len(argc) == 4:
            tri = argc[0]
            argc = argc[1:]
        else:
            tri = None
        if tri is not None:
            self.setvar('tri', tri)
        return FigSurface.__init__(self, *argc, **kywds)

    @classmethod
    def get_namebase(self):
        return 'trisurface'

    @classmethod
    def property_in_palette(self):
        return ["facecolor_2",
                "edgecolor_2", "linewidthz", "solid_shade",
                "alpha_2"]

    def _args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y", "z"]
        use_np = [True]*3
        names = names + names0
        use_np = use_np + [False]*len(names0)
        values = self.put_args2var(names, use_np)
        x = values[0]
        y = values[1]
        z = values[2]
        if x is None and y is None and z.ndim == 2:
            xtmp = np.arange(z.shape[1])
            ytmp = np.arange(z.shape[0])
            x, y = np.meshgrid(xtmp, ytmp)
            self.setp("x", x)
            self.setp("y", y)
        if x is None and y is None and z.ndim == 1:
            self.setp("x", None)
            self.setp("y", None)
            self.setp("z", z)
            return True
        if x is None:
            return False
        if y is None:
            return False
        if z is None:
            z = np.array([0]*len(x))
        if len(z) == 1:
            z = np.array([z[0]]*len(x))
        self.setp("z", z)
        return True

    def generate_artist(self):
        if self.isempty() is False:
            return

        axes = self.get_figaxes()
        if not axes.get_3d():
            return

        container = self.get_container()
        x, y, z = self._eval_xy()
        # use_var should be false if evaluation is
        # okey.
        if self.getp('use_var'):
            return

        kywds = self._var["kywds"]

        cax = self.get_caxisparam()
        if cax is None:
            dprint1('Error: cax is None')
            return
        # if self.getp('alpha') is not None else 1
        kywds['alpha'] = self.getp('alpha')

        fc = self.getp('facecolor')
        if isinstance(fc, str):
            fc = cc.to_rgba(fc)
        if fc is None:
            fc = [0, 0, 0, 0]
        else:
            fc = list(fc)
            if self.getp('alpha') is not None:
                fc[3] = self.getp('alpha')
        ec = self.getp('edgecolor')
        if isinstance(ec, str):
            ec = cc.to_rgba(ec)
        if ec is None:
            ec = [0, 0, 0, 0]
        else:
            ec = list(ec)
            if self.getp('alpha') is not None:
                ec[3] = self.getp('alpha')
        cz = self.getvar('cz')
        if cz is None:
            kywds['cz'] = False
        else:
            kywds['cz'] = cz
        if kywds['cz']:
            kywds['cdata'] = self.getvar('cdata')
        kywds['facecolor'] = (fc,)
        kywds['edgecolor'] = (ec,)
        kywds['linewidths'] = 0.0 if self.getp(
            'linewidth') is None else self.getp('linewidth')
        kywds['shade'] = self.getvar('shade')

        args, self._tri = tri_args(x, y, self._tri)
        args.append(z.flatten().astype(float))
        # if self.hasvar('tri'):
        #    args = (self.getvar('tri'), z)
        # else:
        #    args = (x, y, z)
        self._artists = [container.plot_trisurf(*args, **kywds)]

        for artist in self._artists:
            artist.do_stencil_test = False
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))
            if self.getvar('cz'):
                cax = self.get_caxisparam()
                if cax is None:
                    dprint1('Error: cax is None')
                cax.set_crangeparam_to_artist(artist)

    def get_xrange(self, xrange=[None, None], scale='linear'):
        if self.hasvar('tri'):
            tri = self.getvar('tri')
            return self._update_range(xrange, (np.nanmin(tri.x), np.nanmax(tri.x)))
        else:
            return super(FigTrisurface, self).get_xrange(xrange=xrange,
                                                         scale='linear')

    def get_yrange(self, yrange=[None, None], xrange=[None, None], scale='linear'):
        if self.hasvar('tri'):
            tri = self.getvar('tri')
            return self._update_range(yrange, (np.nanmin(tri.y), np.nanmax(tri.y)))
        else:
            return super(FigTrisurface, self).get_yrange(yrange=yrange,
                                                         xrange=xrange, scale='linear')

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        if self.hasvar('tri'):
            x, y, z = self._eval_xy()
            return self._update_range(zrange, (np.min(z), np.max(z)))
        else:
            return super(FigTrisurface, self).get_zrange(zrange=zrange,
                                                         xrange=xrange,
                                                         yrange=yrange,
                                                         scale='linear')
    '''    
    def get_crange(self, crange=[None,None], 
                         xrange=[None,None], 
                         yrange=[None,None],
                         scale = 'linear'):
        cdata = self.getvar('cdata')
        cz = self.getvar('cz')
        if not cz: return crange
        if cdata is None:
            x, y, z = self._eval_xy()
            crange = self._update_range(crange,
                             (np.nanmin(z), np.nanmax(z)))

        else:
            if np.iscomplexobj(cdata):
                tmp  = np.max(np.abs(cdata))
                crange = self._update_range(crange,
                                    (-tmp, tmp))

            else:
                crange = self._update_range(crange,
                                    (np.nanmin(cdata), np.nanmax(cdata)))

        return crange

    '''



#
#   fig_scatter
#
#       support scatter plot
#
#       in matplotlib
#       scatter(x, y, s=20, c=u'b', marker=u'o', cmap=None,
#               norm=None, vmin=None, vmax=None, alpha=None,
#               linewidths=None, verts=None, **kwargs)
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
# *******************************************
#     Copyright(c) 2012- PSFC, MIT
# *******************************************

from ifigure.mto.fig_obj import FigObj, mask_negative
from ifigure.mto.axis_user import XUser, YUser, CUser, ZUser
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import sys
import numpy as np
import ifigure.utils.cbook as cbook
import ifigure.widgets.canvas.custom_picker as cpicker
from ifigure.utils.cbook import ProcessKeywords
from matplotlib.cm import ScalarMappable
from matplotlib.patches import Rectangle
from matplotlib.transforms import Bbox
from ifigure.utils.args_parser import ArgsParser
from matplotlib.colors import Colormap

#
#  debug setting
#
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('FigScatter')

default_kargs = {}


class FigScatter(FigObj, XUser, YUser, ZUser, CUser):
    def __new__(cls, *args, **kywds):
        def set_hidden_vars(obj):
            obj._tri = None
            for key in default_kargs:
                if not obj.hasp(key):
                    obj.setp(key, default_kargs[key])
                if not obj.hasvar(key):
                    obj.setvar(key, default_kargs[key])
            return obj

        if 'src' in kywds:
            obj = FigObj.__new__(cls, *args, **kywds)
            obj = set_hidden_vars(obj)
            return obj

        p = ArgsParser()
        p.add_var('x', ['numbers|nonstr', 'dynamic'])
        p.add_var('y', ['numbers|nonstr', 'dynamic'])
        p.add_opt('z', None, ['numbers|nonstr'])
        p.add_key('s', 20)
        p.add_key('c', 'b', ['numbers', 'str'])
        p.add_key('marker', 'o')
        p.add_key('sscale', 1.0)
        p.set_ndconvert("x", "y", "z",)
        p.set_squeeze_minimum_1D("x", "y", "z", )
#        p.set_default_list(default_kargs)
        p.add_key('alpha', 1.0)
        p.add_key('cmap', 'jet')

        v, kywds, d, flag = p.process(*args, **kywds)
        if not flag:
            raise ValueError('Failed when processing argument')

        obj = FigObj.__new__(cls, *args, **kywds)
        obj = set_hidden_vars(obj)

        for name in ('x', 'y', 'z', 's', 'c', 'alpha',  'marker', 'sscale'):
            obj.setvar(name, v[name])
        if v['cmap'] is not None:
            if isinstance(v['cmap'], Colormap):
                v['cmap'] = v['cmap'].name
            kywds['cmap'] = v['cmap']
            del v['cmap']
        obj.setvar("kywds", kywds)

        return obj

    def __init__(self, *args, **kywds):
        self._data_extent = None
        XUser.__init__(self)
        YUser.__init__(self)
        ZUser.__init__(self)
        CUser.__init__(self)

        self._pick_pos = None
        self._cb_added = False
        args = []
        if 'src' not in kywds:
            kywds = self.getvar("kywds")
        super(FigScatter, self).__init__(**kywds)

    @classmethod
    def isFigScatter(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'scatter'

    @classmethod
    def property_in_file(self):
        return (["linewidth"] +
                super(FigScatter, self).property_in_file())

    @classmethod
    def property_in_palette(self):
        return ["marker_2", "linewidthz", "scatter_sscale", "alpha_2"]

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ([] +
                super(FigScatter, self).attr_in_file())

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'image.png')
        return [idx1]

    def set_parent(self, parent):
        XUser.unset_ax(self)
        YUser.unset_ay(self)
        ZUser.unset_az(self)
        CUser.unset_ac(self)
        super(FigScatter, self).set_parent(parent)
        XUser.get_xaxisparam(self)
        YUser.get_yaxisparam(self)
        ZUser.get_zaxisparam(self)
        CUser.get_caxisparam(self)

    def args2var(self):
        ret = self._args2var()
        if ret:
            if 'cmap' in self.getvar("kywds"):
                cax = self.get_caxisparam()
                cax.set_cmap(self.getvar('kywds')['cmap'])
        return ret

    def _args2var(self):
        names0 = self.attr_in_file()
        names = ["x", "y", "z"] + names0
        use_np = [True]*3 + [False]*len(names0)
        values = self.put_args2var(names,
                                   use_np)
        x, y, z, = values[:3]
        if x is None:
            return False
        if y is None:
            return False

        return True

    def generate_artist(self):
        # this method generate artist
        # if artist does exist, update artist
        # based on the information specifically
        # managed by fig_obj tree. Any property
        # internally managed by matplotlib
        # does not change
        container = self.get_container()
        if container is None:
            return
#           if self.get_figaxes().get_3d(): return
        if self.isempty() is False:
            return
        x, y, z = self._eval_xyz()
        lp = self.getp("loaded_property")

        cax = self.get_caxisparam()
        if cax is None:
            dprint1('Error: cax is None')
            return
        crange = cax.range

        if lp is None or len(lp) == 0:
            if self.get_figaxes().get_3d():
                args = (x, y, z)
                extent = (np.min(x), np.max(x),
                          np.min(y), np.max(y), )
            else:
                args = (x, y)
                extent = (self.get_xaxisparam().range +
                          self.get_yaxisparam().range)

            kywds = self._var["kywds"].copy()
            kywds['alpha'] = self.getvar('alpha')
            kywds['s'] = self.getvar('s')*self.getvar('sscale')
            kywds['c'] = self.getvar('c')
            kywds['marker'] = self.getvar('marker')
            if cax.scale == 'linear':
                kywds["vmin"] = crange[0]
                kywds["vmax"] = crange[1]
            else:
                # args.append(np.log10(zp))
                kywds["vmin"] = np.log10(max((crange[0], 0)))
                kywds["vmax"] = np.log10(max((crange[1], 0)))

            self.set_artist(container.scatter(*args,
                                              **kywds))
            cax.set_crangeparam_to_artist(self._artists[0])
        else:
            x, y, z = self.getp(('x', 'y', 'z'))
            if self.get_figaxes().get_3d():
                args = (x, y, z)
                extent = (np.min(x), np.max(x),
                          np.min(y), np.max(y), )
            else:
                args = (x, y,)
                extent = (self.get_xaxisparam().range +
                          self.get_yaxisparam().range)

            kywds = lp[0]
            kywds['s'] = self.getvar('s')*self.getvar('sscale')
            kywds['c'] = self.getvar('c')
            kywds['marker'] = self.getvar('marker')
            self.set_artist(container.scatter(*args,
                                              **kywds))
            cax.set_crangeparam_to_artist(self._artists[0])
        self.delp("loaded_property")

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = self._artists

        self.store_loaded_property()

#        self.highlight_artist(False, artistlist)
        for a in artistlist:
            a.remove()

        super(FigScatter, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        alist = self._artists
        container = self.get_container()
        if val == True:
            x, y, z = self.getp(('x', 'y', 'z'))
            if self.get_figaxes().get_3d():
                args = (x, y, z)
            else:
                args = (x, y)
            kwargs = {'marker': 's',
                      'color': 'k',
                      'linestyle': 'None',
                      'markerfacecolor': 'None',
                      'markeredgewidth': 0.5,
                      'scalex': False, 'scaley': False}

            hl = container.plot(*args, **kwargs)
            for a in alist:
                for item in hl:
                    a.figobj_hl.append(item)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    hl.remove()
                a.figobj_hl = []

    def picker_a(self, artist, evt):
        hit, extra = artist.contains(evt)
        if hit:
            return True,  {'child_artist': artist}
        else:
            return False, {}

    def set_marker(self, value, a):
        if self.getvar('marker') == value:
            return
        self.setvar('marker', value)
        if len(self._artists) > 0:
            a1 = self._artists[0]
            self.refresh_artist_data()
            a2 = self._artists[0]
            ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def get_marker(self, a):
        return self.getvar('marker')

    def set_sscale(self, value, a):
        if self.getvar('sscale') == value:
            return

        self.setvar('sscale', value)
        if len(self._artists) > 0:
            a1 = self._artists[0]
            self.refresh_artist_data()
            a2 = self._artists[0]
            ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def get_sscale(self, a):
        return self.getvar('sscale')

    def set_linewidth(self, value, a):
        v = [value for x in self._artists[0].get_linewidth()]
        self._artists[0].set_linewidth(tuple(v))

    def get_linewidth(self, value):
        v = self._artists[0].get_linewidth()
        if len(v) > 1:
            return None
        else:
            return v[0]

    def set_alpha(self, value, a):
        self.setvar('alpha', value)
        fc = self._artists[0].get_facecolor()
        fc = [(x[0], x[1], x[2], value) for x in fc]
        self._artists[0].set_facecolor(tuple(fc))

    def get_alpha(self, a):
        return self.getvar('alpha')

    def get_mappable(self):
        return [a for a in self._artists if isinstance(a, ScalarMappable)]

    def get_xrange(self, xrange=[None, None], scale='linear'):
        x, y, z = self._eval_xyz()
        if x is None:
            return xrange
        if scale == 'log':
            x = mask_negative(x)
        return self._update_range(xrange, [np.nanmin(x), np.nanmax(x)])

    def get_yrange(self, yrange=[None, None],
                   xrange=[None, None], scale='linear'):
        #        de = self.get_data_extent()
        x, y, z = self._eval_xyz()
        if x is None:
            return yrange
        if y is None:
            return yrange
        if scale == 'log':
            y = mask_negative(y)
        return self._update_range(yrange, (np.nanmin(y), np.nanmax(y)))

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        x, y, z = self._eval_xyz()
        c = self.getvar('c')
        if isinstance(c, str):
            return crange
        if hasattr(c, 'ndim'):
            if c.ndim > 1:
                return crange
        return self._update_range(crange, (np.nanmin(c), np.nanmax(c)))

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        if self.get_figaxes().get_3d():
            x, y, z = self._eval_xyz()
            return self._update_range(zrange, (np.nanmin(z), np.nanmax(z)))
        else:
            return zrange

    def _eval_xyz(self):
        if self.getp('use_var'):
            success = self.handle_use_var()
            if not success:
                return None, None, None
        return self.getp(("x", "y", "z"))

    def _saveload_names(self):
        return ['x', 'y', 'z']

    def save_data2(self, data=None):
        def check(obj, name):
            if not isinstance(obj.getp(name), np.ndarray):
                return False
            if not isinstance(obj.getvar(name), np.ndarray):
                return False
            return obj.getp(name) is obj.getvar(name)

        if data is None:
            data = {}
        names = self._saveload_names()
        var = {name: check(self, name) for name in names}
        for name in names:
            if not var[name]:
                var[name+'data'] = self.getp(name)
        data['FigScatter'] = (1, var)
        data = super(FigScatter, self).save_data2(data)

        return data

    def load_data2(self, data):
        d = data['FigScatter']
        super(FigScatter, self).load_data2(data)
        var = d[1]
        names = self._saveload_names()
        for name in names:
            if not name in var:
                var[name] = True
            if var[name]:
                self.setp(name,
                          self.getvar(name))
            else:
                self.setp(name, var[name+'data'])



import wx.stc as stc


def set_python_style(obj):
    from ifigure.widgets.script_editor import faces
    # my own style....
    # Global default styles for all languages
    # self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "fore:#440000,face:%(mono)s,size:%(size)d" % faces)
    # Default
    obj.StyleSetSpec(stc.STC_P_DEFAULT,
                     "fore:#0000ff,face:%(helv)s,size:%(size)d" % faces)
    obj.StyleClearAll()  # Reset all to be like the default
    # Comments
    obj.StyleSetSpec(stc.STC_P_COMMENTLINE,
                     "fore:#007F00,size:%(size)d" % faces)
    # Number
    obj.StyleSetSpec(stc.STC_P_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
    # String
    obj.StyleSetSpec(stc.STC_P_STRING, "fore:#7F007F,size:%(size)d" % faces)
    # Single quoted string
    obj.StyleSetSpec(stc.STC_P_CHARACTER, "fore:#7F007F,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(stc.STC_P_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
    # Triple quotes
    obj.StyleSetSpec(stc.STC_P_TRIPLE, "fore:#7F0000,size:%(size)d" % faces)
    # Triple double quotes
    obj.StyleSetSpec(stc.STC_P_TRIPLEDOUBLE,
                     "fore:#7F0000,size:%(size)d" % faces)
    # Class name definition
    obj.StyleSetSpec(stc.STC_P_CLASSNAME,
                     "fore:#0000FF,bold,underline,size:%(size)d" % faces)
    # Function or method name definition
    obj.StyleSetSpec(stc.STC_P_DEFNAME,
                     "fore:#007F7F,bold,size:%(size)d" % faces)
    # Operators
    obj.StyleSetSpec(stc.STC_P_OPERATOR, "bold,size:%(size)d" % faces)
    # Identifiers
    obj.StyleSetSpec(stc.STC_P_IDENTIFIER,
                     "fore:#000000,size:%(size)d" % faces)
    # Comment-blocks
    obj.StyleSetSpec(stc.STC_P_COMMENTBLOCK,
                     "fore:#7F7F7F,size:%(size)d" % faces)
    # End of line where string is not closed
    obj.StyleSetSpec(stc.STC_P_STRINGEOL,
                     "fore:#000000,back:#E0C0E0,eol,size:%(size)d" % faces)


def set_cpp_style(obj):
    from script_editor import faces
    # my own style....
    # Global default styles for all languages
    # self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "fore:#440000,face:%(mono)s,size:%(size)d" % faces)
    # Default
    obj.StyleSetSpec(stc.STC_C_DEFAULT,
                     "fore:#0000ff,face:%(helv)s,size:%(size)d" % faces)
    obj.StyleClearAll()  # Reset all to be like the default
    # Comments
    obj.StyleSetSpec(stc.STC_C_COMMENTLINE,
                     "fore:#007F00,size:%(size)d" % faces)
    # Number
    obj.StyleSetSpec(stc.STC_C_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
    # String
    obj.StyleSetSpec(stc.STC_C_STRING, "fore:#7F007F,size:%(size)d" % faces)
    # Single quoted string
    obj.StyleSetSpec(stc.STC_C_CHARACTER, "fore:#7F007F,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(stc.STC_C_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
    # Class name definition
    obj.StyleSetSpec(stc.STC_C_GLOBALCLASS,
                     "fore:#0000FF,bold,underline,size:%(size)d" % faces)
    # Operators
    obj.StyleSetSpec(stc.STC_C_OPERATOR, "bold,size:%(size)d" % faces)
    # Identifiers
    obj.StyleSetSpec(stc.STC_C_IDENTIFIER,
                     "fore:#000000,size:%(size)d" % faces)
    # Comment-blocks
    obj.StyleSetSpec(stc.STC_C_COMMENTDOC,
                     "fore:#00F700,size:%(size)d" % faces)
    # Comment-blocks
    obj.StyleSetSpec(stc.STC_C_COMMENTDOCKEYWORD,
                     "fore:#007F00,size:%(size)d" % faces)
    # Comment-blocks
    obj.StyleSetSpec(stc.STC_C_COMMENTDOCKEYWORDERROR, "fore:#007F00,size:%(size)d"
                     % faces)
    # End of line where string is not closed
    obj.StyleSetSpec(stc.STC_P_STRINGEOL,
                     "fore:#000000,back:#E0C0E0,eol,size:%(size)d" % faces)


def set_f77_style(obj):
    from ifigure.widgets.script_editor import faces
    # my own style....
    # Global default styles for all languages
    # self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "fore:#440000,face:%(mono)s,size:%(size)d" % faces)
    # Default
    obj.StyleSetSpec(stc.STC_F_DEFAULT,
                     "fore:#0000ff,face:%(helv)s,size:%(size)d" % faces)
    obj.StyleClearAll()  # Reset all to be like the default
    # Comments
    obj.StyleSetSpec(stc.STC_F_COMMENT, "fore:#007F00,size:%(size)d" % faces)
    # Number
    obj.StyleSetSpec(stc.STC_F_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
    # String
    obj.StyleSetSpec(stc.STC_F_STRING1, "fore:#7F007F,size:%(size)d" % faces)
    # String
    obj.StyleSetSpec(stc.STC_F_STRING2, "fore:#7F007F,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(stc.STC_F_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(
        stc.STC_F_WORD2, "fore:#00007F,bold,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(
        stc.STC_F_WORD3, "fore:#00007F,bold,size:%(size)d" % faces)
    # Preprocessor
    obj.StyleSetSpec(stc.STC_F_PREPROCESSOR,
                     "fore:#7F7F7F,size:%(size)d" % faces)
    # Operators
    obj.StyleSetSpec(stc.STC_F_OPERATOR, "bold,size:%(size)d" % faces)
    # Operators
    obj.StyleSetSpec(stc.STC_F_OPERATOR2, "bold,size:%(size)d" % faces)
    # Identifiers
    obj.StyleSetSpec(stc.STC_F_IDENTIFIER,
                     "fore:#000000,size:%(size)d" % faces)
    # End of line where string is not closed
    obj.StyleSetSpec(stc.STC_F_STRINGEOL,
                     "fore:#000000,back:#E0C0E0,eol,size:%(size)d" % faces)


def set_fortran_style(obj):
    from ifigure.widgets.script_editor import faces
    # my own style....
    # Global default styles for all languages
    # self.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "fore:#440000,face:%(mono)s,size:%(size)d" % faces)
    # Default
    obj.StyleSetSpec(stc.STC_F_DEFAULT,
                     "fore:#0000ff,face:%(helv)s,size:%(size)d" % faces)
    obj.StyleClearAll()  # Reset all to be like the default
    # Comments
    obj.StyleSetSpec(stc.STC_F_COMMENT, "fore:#007F00,size:%(size)d" % faces)
    # Number
    obj.StyleSetSpec(stc.STC_F_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
    # String
    obj.StyleSetSpec(stc.STC_F_STRING1, "fore:#7F007F,size:%(size)d" % faces)
    # String
    obj.StyleSetSpec(stc.STC_F_STRING2, "fore:#7F007F,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(stc.STC_F_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(
        stc.STC_F_WORD2, "fore:#00007F,bold,size:%(size)d" % faces)
    # Keyword
    obj.StyleSetSpec(
        stc.STC_F_WORD3, "fore:#00007F,bold,size:%(size)d" % faces)
    # Preprocessor
    obj.StyleSetSpec(stc.STC_F_PREPROCESSOR,
                     "fore:#7F7F7F,size:%(size)d" % faces)
    # Operators
    obj.StyleSetSpec(stc.STC_F_OPERATOR, "bold,size:%(size)d" % faces)
    # Operators
    obj.StyleSetSpec(stc.STC_F_OPERATOR2, "bold,size:%(size)d" % faces)
    # Identifiers
    obj.StyleSetSpec(stc.STC_F_IDENTIFIER,
                     "fore:#000000,size:%(size)d" % faces)
    # End of line where string is not closed
    obj.StyleSetSpec(stc.STC_F_STRINGEOL,
                     "fore:#000000,back:#E0C0E0,eol,size:%(size)d" % faces)



# uncomment the following to use wx rather than wxagg
from matplotlib.backends.backend_agg import RendererAgg
from ifigure.matplotlib_mod.backend_wxagg_mod import FigureCanvasWxAggMod
from distutils.version import LooseVersion
import numpy as np
import time
import wx

import weakref
import array

from ctypes import sizeof, c_float, c_void_p, c_uint, string_at

import matplotlib

from functools import wraps
from matplotlib.backends.backend_wx import FigureCanvasWx as Canvas
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as CanvasAgg
from matplotlib.backends.backend_wx import RendererWx
from ifigure.utils.cbook import EraseBitMap
from operator import itemgetter

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('BackendWXAggGL')

isMPL_before_1_2 = LooseVersion(matplotlib.__version__) < LooseVersion("1.2")

use_gl_12 = True


def load_glcanvas(debug=False):
    if wx.__version__[0] == '4' and not use_gl_12:
        from ifigure.matplotlib_mod.glcanvas_15 import MyGLCanvas
        if debug:
            dprint1('loading glcanvas15')
    else:
        from ifigure.matplotlib_mod.glcanvas_12 import MyGLCanvas
        if debug:
            dprint1('loading glcanvas12')
    return MyGLCanvas


#from renderer_gl import RendererGL


class RendererGLMixin(object):
    def __init__(self, *args, **kwargs):
        self.do_stencil_test = False
        self._no_update_id = False
        self._num_globj = 0

    def __del__(self):
        self._glcanvas = None

#    def set_glcanvas(self, glcanvas):
#        self._glcanvas = glcanvas

    def get_glcanvas(self):
        from ifigure.matplotlib_mod.backend_wxagg_gl import FigureCanvasWxAggModGL
        return FigureCanvasWxAggModGL.glcanvas

    def gl_draw_image(self, gc, path, trans, im, **kwargs):
        self.get_glcanvas().store_draw_request('image', gc,
                                               path, trans,
                                               im, **kwargs)

    def gl_draw_markers(self, gc, marker_path, marker_trans,
                        path, transform, rgbFace=None, **kwargs):
        self.get_glcanvas().store_draw_request('markers', gc, marker_path, marker_trans,
                                               path, transform,  **kwargs)

    def gl_draw_path(self, gc, path, transform, rgbFace=None, rgbEdge=None, **kwargs):
        self.get_glcanvas().store_draw_request('path', gc, path, rgbFace, rgbEdge, **kwargs)

    def gl_draw_path_collection(self, gc, frozen, paths,
                                transform, offset, transOffset,
                                facecolor, edgecolor,
                                linewidth, linestyle,
                                antialias, urls, offset_position, **kwargs):

        self.get_glcanvas().store_draw_request('path_collection', gc, paths,
                                               facecolor, edgecolor,
                                               linewidth, linestyle, offset,
                                               **kwargs)

    def gl_draw_path_collection_e(self, gc, frozen, paths,
                                  transform, offset, transOffset,
                                  facecolor, edgecolor,
                                  linewidth, linestyle,
                                  antialias, urls, offset_position, **kwargs):
        self.get_glcanvas().store_draw_request('path_collection_e', gc, paths,
                                               facecolor, edgecolor,
                                               linewidth, linestyle, offset,
                                               **kwargs)

#    def set_canvas_proj(self, worldM, viewM, perspM, lookat):
#        self.get_glcanvas().set_proj(worldM, viewM, perspM, lookat)

    def update_id_data(self, data, tag=None):
        if not self._no_update_id:
            tag._gl_id_data = data
        self._no_update_id = False

    def no_update_id(self):
        self._no_update_id = True


def mixin_gl_renderer(renderer):
    if hasattr(renderer, '_gl_renderer'):
        return
    renderer._gl_renderer = RendererGLMixin()
    renderer.gl_svg_rescale = False
    names = ('gl_draw_markers', 'gl_draw_path_collection',
             'gl_draw_path_collection_e',
             'gl_draw_path', 'gl_draw_image',
             'get_glcanvas',
             'update_id_data', 'no_update_id')
    for name in names:
        setattr(renderer, name, getattr(renderer._gl_renderer, name))


class FigureCanvasWxAggModGL(FigureCanvasWxAggMod):
    glcanvas = None

    def __init__(self, *args, **kwargs):
        FigureCanvasWxAggMod.__init__(self, *args, **kwargs)
        if FigureCanvasWxAggModGL.glcanvas is None:
            win = wx.GetApp().TopWindow
            glcanvas = load_glcanvas()(win)
            FigureCanvasWxAggModGL.glcanvas = glcanvas
            glcanvas.SetMinSize((2, 2))
            glcanvas.SetMaxSize((2, 2))
            # self.SetSizer(wx.BoxSizer(wx.HORIZONTAL))
            win.GetSizer().Add(glcanvas)
            win.Layout()
            glcanvas.Refresh()

    def get_renderer(self, cleared=False):
        l, b, w, h = self.figure.bbox.bounds
        key = w, h, self.figure.dpi
        try:
            self._lastKey, self.renderer
        except AttributeError:
            need_new_renderer = True
        else:
            need_new_renderer = (self._lastKey != key)

        if need_new_renderer:
            self.renderer = RendererAgg(w, h, self.figure.dpi)
            mixin_gl_renderer(self.renderer)
            self._lastKey = key
        elif cleared:
            self.renderer.clear()
        return self.renderer

    def draw(self, *args, **kargs):
        self._update_hl_color()
        return FigureCanvasWxAggMod.draw(self,  *args, **kargs)

    def draw_artist(self, drawDC=None, alist=None):
        if alist is None:
            alist = []
        gl_obj = [a for a in alist if hasattr(a, 'is_gl')]

        for o in gl_obj:
            o.is_last = False
        if len(gl_obj) > 0:
            gl_obj[-1].is_last = True
            self.renderer._k_globj = 0
            self.renderer._num_globj = len(gl_obj)
            self.renderer.no_update_id()
#            self.renderer.no_lighting = no_lighting
            self._update_hl_color()
            FigureCanvasWxAggModGL.glcanvas._artist_mask = alist

        v = FigureCanvasWxAggMod.draw_artist(self, drawDC=drawDC, alist=alist)
#        self.renderer.no_lighting = False
        return v

    def _update_hl_color(self):
        value = self.hl_color
        vv = list([float(x) for x in value]) + [1.0]
        vv[3] = 0.65
        FigureCanvasWxAggModGL.glcanvas._hl_color = tuple(vv[:4])

    def _onPaint(self, evt):
        #        self.glcanvas.OnPaint(evt)
        #        evt.Skip()
        FigureCanvasWxAggMod._onPaint(self, evt)

    def _onSize(self, evt=None, nocheck=False):
        FigureCanvasWxAggModGL.glcanvas._hittest_map_update = True
        FigureCanvasWxAggMod._onSize(self, evt=evt, nocheck=nocheck)
        # self.glcanvas.SetSize(self.bitmap.GetSize())

    def disable_alpha_blend(self):
        FigureCanvasWxAggModGL.glcanvas._alpha_blend = False

    def enable_alpha_blend(self):
        FigureCanvasWxAggModGL.glcanvas._alpha_blend = True



from __future__ import print_function
#  Name   :fig_obj
#
#          base class for fig_obj
#
#
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History :
#
# *******************************************
#     Copyright(c) 2012- S. Shiraiwa
# *******************************************
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.ifigure_config import isMPL2
from ifigure.widgets.canvas.file_structure import *

from ifigure.mto.treedict import TreeDict
#from ifigure.mto.py_script import PyScript
from ifigure.mto.py_code import PyCode
from ifigure.mto.axis_user import XUser, YUser, ZUser, CUser
from ifigure.mto.metadata import MetadataHolder
import ifigure.utils.geom as geom
import ifigure.events
import logging
import weakref
import traceback
import ifigure.utils.pickle_wrapper as pickle
from matplotlib.lines import Line2D
from ifigure.utils.cbook import isiterable
import numpy as np

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('FigObj')


def set_mpl_all(artists, prop, value):
    for a in artists:
        m = getattr(a, 'set_'+prop)
        m(value)


def get_mpl_first(artists, prop):
    a = artists[0]
    m = getattr(a, 'get_'+prop)
    return m()


def properties_in_file_0(obj):
    from ifigure.widgets.artist_widgets import listparam
    props = obj._property_in_file(obj)
    return [listparam[p][4] for p in prop if listparam[p][5] == 0]


def properties_in_file_1(obj):
    from ifigure.widgets.artist_widgets import listparam
    props = obj._property_in_file(obj)
    return [listparam[p][4] for p in prop if listparam[p][5] == 1]


def properties_in_file_2(obj):
    from ifigure.widgets.artist_widgets import listparam
    props = obj._property_in_file(obj)
    return [listparam[p][4] for p in prop if (listparam[p][5] == 12
                                              or listparam[p][5] == 2)]


def properties_in_file_3(obj):
    from ifigure.widgets.artist_widgets import listparam
    props = obj._property_in_file(obj)
    return [listparam[p][4] for p in prop if (listparam[p][5] == 13
                                              or listparam[p][5] == 3)]


def mask_negative(z):
    z_masked = np.ma.masked_array(z)
    z_masked[z <= 0] = np.ma.masked
    return z_masked

# class FigObj(PyScript):


class FigObj(TreeDict, MetadataHolder):
    '''
    FigObj:
       baseclass of all figure object

    '''
    default_rasterized = False

    def __new__(cls, *args, **kargs):
        obj = TreeDict.__new__(cls, *args, **kargs)

        # attr variable is used to control
        # matplotlib, and intended not to seen
        # by users

        obj._attr = {}
        obj._artists = []
        obj._mappable = None
        obj._a_update = False
        obj._update_client = []
        obj._isPageObj = False
        # _axes is a weak reference to fig_axes
        obj._fig_axes = []
        obj._picker_a_type = 'area'
        obj._picker_a_loc = 3
        obj._picker_a_mode = 0
        obj._pickmask = False
        obj._drag_hl = None
        obj._bmp = None
        obj._bmp_update = False
        obj._container_idx = 0
        obj._data_extent = None

        obj._floating = False
        # True if obj is floating on axes
        # used for inset_axes, colorbar, legend
        obj._cursor1 = []
        obj._cursor1_a = None
        obj._cursor2 = []
        obj._cursor2_a = None
        obj._cursor_range = (-np.inf, np.inf)
        obj._cursor_data = [tuple(), tuple(), tuple()]
        obj._eval_fifo_length = 100
        obj._eval_mode = 'replace'
        return obj

#    def __repr__(self):
#        return self.get_full_path()

    def __init__(self, *args, **kywds):
        #        self._container=None

        kywds, parent = self._process_kywd2(kywds, 'parent', None)
        kywds, src = self._process_kywd2(kywds, 'src', None)
        kywds, zorder = self._process_kywd2(kywds, 'zorder', 1)
        kywds, rasterized = self._process_kywd2(kywds, 'rasterized',
                                                self.default_rasterized)
        self._show_scriptmenu = False

        super(FigObj, self).__init__(parent=parent, src=src)
        self._has_private_owndir = True
        #   at this point all arguments should have
        #   been processed. Here, the left over is
        #   kept and will be simply passed to
        #   artist generator
        self.setp('use_var', True)
        self.setp('zorder', zorder)
        self.setp('rasterized', rasterized)
        self.setp('frameart', False)
        self.setp('noclip3d', False)
        self.setp("args", args)
        if len(args) != 0:
            logging.warning(
                "FigObj:len(args) != 0: Not all arguments may not have processed")
            print(self)
            print(args)
        #  Properties of figobj have different
        #  level of complexity
        #
        #  1) some mpl routins return an artist,
        #   editting those property makes sense.
        #   for such an property,
        #     1) palette edit artist property
        #        directly
        #     2) cut/paste, save/load read these
        #        property from artists
        #     3) keyword unsupported by figobj
        #        is stored in _var and used to
        #        generate artist first time.
        #
        #
        #  2) Other mpl routine need a special care.
        #   For example, contour returns patchs.
        #   it is not possible to edit these patch
        #   properteis to change number of levels.
        #   In such case, generate_artist needs
        #   to call mpl everytime when those properties
        #   are changed.

        #
        #     1) palette edit _attr in figobj. these
        #        are set when artist are generated.
        #
        #     2) cut/paste, save/load read these
        #        property from artists
        #

        #  3) and most complicated case is when an
        #      figobj needs to handle both. this case
        #      is not yet implemented.
        #      Switching from hotizontal to vertical
        #      line in axsplan/axline will be such a
        #      case

        #   keyword argments handled by figobj is
        #   split from the list.
        #   the rest are stored and will be passed
        #   artist generator
        #   extracted keywords are copyed to _attr
        #   when an atrist is generated, and can
        #   be edited from palette.

        self.setvar("kywds", kywds)

    @classmethod
    def isFigObj(self):
        return True

    def isPageObj(self):
        return self._isPageObj

    @classmethod
    def get_namebase(self):
        return 'figobj'

    @classmethod
    def allow_outside(self):
        ### if fig_obj can be dragged outside of axes ###
        return False

    @classmethod
    def property_in_file(self):
        # define artist property read by mpl.getp and saved in file
        return ['zorder', 'rasterized']
#    @classmethod
#    def _property_in_file(self):
#        ###define artist property read by mpl.getp and saved in file
#        return ['zorder', 'use_var', 'rasterized']

    @classmethod
    def property_in_palette(self):
        # define artist property or _attr shown in palette
        # this function retuns list of key in listparam defined in
        # artist_widgets.py. the 4th field listparam is the name
        # of property read by either mpl.getp (from artist) or
        # figobj.getp (from figobj)
        return []

    @classmethod
    def property_in_palette_axes(self):
        # define artist property or _attr shown
        # in the second panel (axes panel) of property editor
        # the format of return value is the same
        # as property_in_palette(self)
        return []

    @classmethod
    def property_for_shell(self):
        # define artist property which can be edited from shell
        # by default, this returns the same thing as property_in_palette
        # fig_page and fig_axes needs to implement this differently
        # since section editor, axes editor is different from artist_editor
        return self.property_in_palette()

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ['zorder', 'use_var', 'rasterized', 'frameart', 'noclip3d']

    @classmethod
    def can_have_child(self, child=None):
        return isinstance(child, FigObj)

    @classmethod
    def isCompound(cls):
        '''
        Compount FigObj is a type of FigObj which has a mulltiple
        pickable objects. This is introduced to support GL object
        using array_idx. See FigCompound
        '''
        return False

    def isSelected(self):
        '''
        Compound FigObj could return False
        '''
        return True

    def var2p(self, list):
        '''

        '''
        print('FigObj::var2p is obsolete, use put_args2var')
        a = {}
        for name in list:
            if not self.hasvar(name):
                continue
            v = self.eval(name)
            self.setp(name, v)
            a[name] = v
        return a

    def put_args2var(self, names, use_np):
        '''
        a utility to perform eval and put value
        to _attr (setp)
        '''
        ans = []
        k = 0
        for name in names:
            if not self.hasvar(name):
                ans.append(None)
                continue
            else:
                v = self.eval(name, np=use_np[k])
                self.setp(name, v)
                ans.append(v)
            k = k+1
        return tuple(ans)

    def handle_use_var(self):
        '''
        args2var should be implemented in subclass.
        used to gether with put_args2var, subclass
        needs to specify the name of values and
        to handle exceptional cases.
        '''
        if self.getp('use_var'):
            success = self.args2var()
            if success:
                self.setp("use_var", False)
                return True
            else:
                return False

    def process_kywd(self, kywds, name, d):
        '''
        process keyword given to __init__
        store them to var
        '''
        if name in kywds:
            self.setvar(name, kywds[name])
            del kywds[name]
        else:
            self.setvar(name, d)
        return kywds

    def get_container(self, *args, **kywds):
        '''
        this assumes that the parent object has
        only one artist and my artist is child of
        it. if it is not the case, this should
        be overwritten
        '''
        ax = self.get_figaxes()
        if ax is None or ax is self:
            page = self.get_figpage()
            if len(page._artists) == 0:
                return None
            return page._artists[0]
        else:
            if len(ax._artists) == 0:
                return None
            if len(ax._artists) > self._container_idx:
                return ax._artists[self._container_idx]
            else:
                return ax._artists[0]

    def set_parent(self, parent):
        from ifigure.mto.fig_page import FigPage
        from ifigure.mto.fig_axes import FigAxes
        super(FigObj, self).set_parent(parent)
        while parent != None:
            if isinstance(parent, FigPage):
                self._isPageObj = True
                return
            if isinstance(parent, FigAxes):
                self._isPageObj = False
                return
            parent = parent.get_parent()

    def get_mappable(self):
        '''
        return a list of scalar mappable artist
        '''
        return []

    def set_bmp(self, bmp, x, y):
        self._bmp = (bmp, x, y)
        self._bmp_update = True

    def get_bmp(self):
        return self._bmp

    def get_bmp_update(self):
        # true if bitmap is updated
        return self._bmp_update

    def set_bmp_update(self, value):
        # bmp_update should set figaxes's _bmp_update
        obj = self.get_figaxes()
        if obj is not None:
            #           if obj.get_full_path() == 'proj.shot_summary.page1.axes2':
            #               traceback.print_stack()
            obj._bmp_update = value

    def canvas_selected(self):
        pass

    def canvas_unselected(self):
        pass

    def isDraggable(self):
        return False

    def destroy(self, clean_owndir=True):
        # destory tree is more complicated than making it.
        ### idea is...
        # 1) walking through the tree is done in treedict
        # 2) fig_obj takes care of self._artists and
        # artists.figobj. this was done by calleing
        # del_artists()
        # 3) derived classes takes care of the relation
        # between an artists and its container.
        # this way,derived class does not need to impliment
        # its destory by default.(del_artist() could have been
        # mergined in destroy(), it is done for the symmetry
        # of having generate_artist() and del_artist()
        self.set_bmp_update(False)
        if self._drag_hl is not None:
            self._drag_hl.figure.lines.remove(self._drag_hl)
            self._drag_hl = None
        self.del_artist(delall=True)
        for key in self._attr:
            self._attr[key] = None
        # calles super class which kills children first
        super(FigObj, self).destroy(clean_owndir=clean_owndir)

    def generate_artist(self, *args, **kywds):
        dprint1("generate_artist should have be overwrittne", self)

        # self._container=container
    def set_artist(self, alist0):
        if isiterable(alist0):
            alist = alist0
        else:
            alist = [alist0]
        self._artists = alist
        for artist in self._artists:
            artist.figobj = self
            artist.figobj_hl = []
            artist.set_zorder(self.getp('zorder'))

    def generate_cursor(self, evt, idx):
        # called during mouse click in cursor mode
        return []

    def reset_cursor_range(self, evt, idx):
        # called for 2D data cursor
        pass

    def update_cursor(self, evt, idx):
        # called during mousedrag in cursor mode
        pass

    def valid_cursor(self):
        # called just before drawing cursor
        return []

    def refresh_artist_data(self):
        # default action when data is changed...
        # redo artist generation
        if not self.isempty():
            self.del_artist(delall=True)
        if self.isempty() and not self._suppress:
            self._data_extent = None  # do I need this???
            self.generate_artist()
        self.set_bmp_update(False)

    def del_artist(self, artist=None, delall=False):
        # this routine isolate figobj and artists
        # before calling this process, artist should be
        # removed from container. this task should be
        # implemented in derived classes

        # note: subclass delete all childrens' artists
        # therefore del_artists need to be paired with
        # realize

        if delall:
            artistlist = self._artists
        else:
            artistlist = artist
        if artistlist is None:
            return

        for a in artistlist:
            a.figobj = None
            a.figobj_hl = []
        b = [a for a in self._artists if artistlist.count(a) == 0]
        self._artists = b

    def reset_artist(self, load_data=None):
        self.del_artist(delall=True)
        self.delp('loaded_property')
        self.setp('use_var', True)
        if hasattr(self, "_data_extent"):
            self._data_extent = None
        if load_data is not None:
            self.load_data2(load_data)
        self.generate_artist()

    def get_artist(self, idx=None):
        if idx is not None:
            return self._artists[idx]
        return self._artists

    def set_pickmask(self, value):
        self._pickmask = value

    def get_artists_for_pick(self):
        return self._artists

    def get_first_artist(self):
        if len(self._artists) == 0:
            return None
        return self._artists[0]

    def walk_artists(self):
        for td in self.walk_tree():
            for a in td._artists:
                yield a

    def walk_allartists(self):
        def walk_children(a):
            yield a
            for a2 in a.get_children():
                for a3 in walk_children(a2):
                    yield a3
        for a in self._artists:
            for x in walk_children(a):
                yield x

    def walk_figobj(self):
        for obj in self.walk_tree():
            if isinstance(obj, FigObj):
                yield obj

    def set_zorder(self, z1, a=None):
        self.setp('zorder', z1)
        for a in self._artists:
            a.set_zorder(z1)
        if len(self._artists) != 0:
            self.set_bmp_update(False)

    def get_zorder(self, a=None):
        return self.getp('zorder')

    def get_artists_for_frameart(self):
        return self._artists

    def set_frameart(self, v, a=None):
        if self.get_figaxes() is not None:
            dprint1("axes object cannot become Frame Artist")
            return
        if self.get_figpage() is None:
            dprint1("object is not placed in page")
            return
        self.setp('frameart', v)
        for a in self.get_artists_for_frameart():
            a._is_frameart = v

        page = self.get_figpage()
        if v:
            page.add_frameart(self)
        else:
            page.rm_frameart(self)

    def get_frameart(self, a=None):
        return self.getp('frameart')

    def swap_zorder(self, z1, z2, get_action=False):
        page = self.get_figpage()
        objs = [figobj for figobj in page.walk_figobj()
                if figobj.getp('zorder') == z2]

        h = [] if get_action else None
        for figobj in objs:
            if get_action:
                h.append(UndoRedoFigobjMethod(
                    figobj._artists[0], 'zorder', z1))
            else:
                figobj.set_zorder(z1)
        if get_action:
            h.append(UndoRedoFigobjMethod(self._artists[0], 'zorder', z2))
        else:
            self.set_zorder(z1)
        return h

    def move_zorder_forward(self, get_action=False):
        z1 = self.getp('zorder')
        z2 = z1+0.001
        return self.swap_zorder(z1, z2, get_action=get_action)

    def move_zorder_backward(self, get_action=False):
        z1 = self.getp('zorder')
        z2 = z1-0.001
        return self.swap_zorder(z1, z2, get_action=get_action)

    def set_zorder_front(self, get_action=False):
        page = self.get_figpage()
        if self.get_frameart():
            l1 = [figobj.getp('zorder') for figobj in page.walk_figobj()
                  if not figobj._floating and figobj.get_frameart()]
        else:
            l1 = [figobj.getp('zorder') for figobj in page.walk_figobj()
                  if not figobj._floating and not figobj.get_frameart()]
        l2 = [a.zorder for a in page.walk_allartists()]

        ret = [] if get_action else None
        if self.get_zorder() <= max(l1):
            # if i am only figobj at front, do nothing
            if l1.count(self.get_zorder()) == 1 and self.get_zorder() == max(l1):
                return ret
            if get_action:
                return [UndoRedoFigobjMethod(self._artists[0], 'zorder', max(l1) + 0.001)]
            else:
                self.set_zorder(max(l1) + 0.001)
        else:
            if get_action:
                return [UndoRedoFigobjMethod(self._artists[0], 'zorder', max(l1+l2) + 0.001)]
            else:
                self.set_zorder(max(l1+l2) + 0.001)

    def set_zorder_bottom(self, get_action=False):
        page = self.get_figpage()
        a = []
        if self.get_frameart():
            l1 = [figobj.getp('zorder') for figobj in page.walk_figobj()
                  if not figobj._floating and figobj.get_frameart()]
        else:
            l1 = [figobj.getp('zorder') for figobj in page.walk_figobj()
                  if not figobj._floating and not figobj.get_frameart()]
        l2 = [a.zorder for a in page.walk_allartists()]
        #print(l1, l2)
        if get_action:
            return [UndoRedoFigobjMethod(self._artists[0], 'zorder', min(l1+l2) - 0.001)]
        else:
            self.set_zorder(min(l1+l2) - 0.001)

    def set_rasterized(self, value=None, a=None):
        '''
        set_rasterized(True or False) : change rasterized and apply it to 
                                        artist
        set_rasterized(): set artist.rasterrized. this is used in 
                          generate_artist or make_artist
        '''

        if value is not None:
            self.setp('rasterized', value)
        value = self.getp('rasterized')
        for a in self._artists:
            a.set_rasterized(value)

    def get_rasterized(self, a=None):
        return self.getp('rasterized')

    def get_rasterized_action(self, value):
        if len(self._artists) == 0:
            return []
        return [UndoRedoFigobjMethod(self._artists[0], 'rasterized', value)]

    def realize(self, realize_gpholder='both'):
        from ifigure.mto.figobj_gpholder import FigObjGPHolder

        do_generate = True

        if realize_gpholder == 'gp' and not isinstance(self, FigObjGPHolder):
            do_generate = False
        if realize_gpholder == 'non_gp' and isinstance(self, FigObjGPHolder):
            do_generate = False

        if not self._suppress:
            if do_generate:
                self.generate_artist()
            for objname, figobj in self.get_children():
                figobj.realize(realize_gpholder=realize_gpholder)
        else:
            if not self.isempty():
                self.del_artist(delall=True)
            for objname, figobj in self.get_children():
                figobj.realize(realize_gpholder=realize_gpholder)

    '''
       artist update...
       it provides the last moment update (just before draw
       call) of artists. it can be useful if an object 
       property depends on other object.
       However, it does not follow the order of depencency,
       and does not guarantee that referred object is 
       already updated. Therefore, the artist update
       should be performed without relying on this whenever
       it is possible
    '''

    def update_artist(self):
        if self._a_update:
            self.do_update_artist()
            self._a_update = False
#        if self._suppress is False:
        for objname, figobj in self.get_children():
            figobj.update_artist()
#       else:
#            for objname, figobj in self.get_children():
#                figobj.update_artist()

    def do_update_artist(self):
        # this should be implemented in derived class
        pass

    def set_update_artist_request(self, *argv, **argk):
        self._a_update = True

    def _clean_update_client_list(self):
        for figobj in self._update_client:
            if figobj() is None:
                self._update_client.remove(figobj)

    def set_client_update_artist_request(self):
        for figobj in self._update_client:
            if figobj() is not None:
                figobj().set_update_artist_request()
            else:
                self._update_client.remove(figobj)

    def add_update_client(self, figobj):
        check = False
        self._clean_update_client_list()
        for figobj2 in self._update_client:
            if figobj2() is figobj:
                return
        self._update_client.append(weakref.ref(figobj))

    def remove_update_client(self, figobj):
        self._clean_update_client_list()
        for figobj2 in self._update_client:
            if figobj2() is figobj:
                self._update_client.remove(figobj2)

    def isempty(self):
        return len(self._artists) == 0

    def highlight_artist(self, val, artist=None):
        pass

    def has_highlight(self, artist):
        return len(artist.figobj_hl) != 0

    def canvas_menu(self):
        if self.get_rasterized():
            return [("Export",  self.onExport, None),
                    ("Copy object path",  self.onCopyPath, None),
                    ("^Rasterized",  self.onUnsetRasterize, None), ]
        else:
            return [("Export",  self.onExport, None),
                    ("Copy object path",  self.onCopyPath, None),
                    ("*Rasterized",  self.onSetRasterize, None), ]

    def onCopyPath(self, evt):
        import wx
        path = self.get_full_path()
        if wx.TheClipboard.Open():
            # This data objects are held by the clipboard,
            # so do not delete them in the app.
            wx.TheClipboard.SetData(wx.TextDataObject(path))
            wx.TheClipboard.Close()

    def onUnsetRasterize(self, evt):
        from ifigure.widgets.undo_redo_history import GlobalHistory
        a = self.get_rasterized_action(False)
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(a, menu_name='unrasteraized')

#        self.set_rasterized(False)

    def onSetRasterize(self, evt):
        from ifigure.widgets.undo_redo_history import GlobalHistory
        a = self.get_rasterized_action(True)
        window = evt.GetEventObject().GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(a, menu_name='rasteraized')

#        self.set_rasterized(True)
    def set_container_idx(self, value, a=None):

        def _set_ax_params(figaxes, param):
            for b in figaxes.get_axes_artist_by_name(param.name):
                param.set_artist_rangeparam(b)
                param.set_tickparam(b, figaxes)

        if isinstance(self, XUser):
            self.unset_ax()
        if isinstance(self, YUser):
            self.unset_ay()
        if isinstance(self, ZUser):
            self.unset_az()
        #if isinstance(self, CUser): self.unset_ac()
        if self.isempty():
            self._container_idx = value
        else:
            self.del_artist(delall=True)
            self._container_idx = value
            self.generate_artist()
        if not self.isPageObj():
            figaxes = self.get_figaxes()
            params = figaxes.get_axis_param_container_idx(value)
            if isinstance(self, XUser) and 'x' in params:
                self.set_ax(params['x'])
                _set_ax_params(figaxes, params['x'])
            if isinstance(self, YUser) and 'y' in params:
                self.set_ay(params['y'])
                _set_ax_params(figaxes, params['y'])
            if isinstance(self, ZUser) and 'z' in params:
                self.set_az(params['z'])
                _set_ax_params(figaxes, params['z'])
            # if isinstance(self, CUser) and 'c' in params:
            #   self.set_ac(params['c'])

    def get_container_idx(self, a=None):
        return self._container_idx

    def set_caxis_idx(self, value, a=None):
        figaxes = self.get_figaxes()
        if figaxes is None:
            return

        if isinstance(self, CUser):
            self.unset_ac()
            if value >= len(figaxes._caxis):
                dprint1('len(self._caxis) is shorter than request...')
                return
            param = figaxes._caxis[value]
            self.set_ac(param)
            for a2 in self.get_mappable():
                param.set_crangeparam_to_artist(a2)

    def get_caxis_idx(self, a=None):
        if self.get_figaxes() is None:
            return None
        return self.get_axis_param_idx()[-1]

    def get_axis_param_idx(self):
        if self.get_figaxes() is None:
            return [None]*4
        fa = self.get_figaxes()
        value = [-1, -1, -1, -1]
        for k, p in enumerate(fa._xaxis):
            if self in [r() for r in p._member]:
                value[0] = k
        for k, p in enumerate(fa._yaxis):
            if self in [r() for r in p._member]:
                value[1] = k
        for k, p in enumerate(fa._zaxis):
            if self in [r() for r in p._member]:
                value[2] = k
        for k, p in enumerate(fa._caxis):
            if self in [r() for r in p._member]:
                value[3] = k
        return value

    def canvas_axes_selection_menu(self, fig_axes):
        def append_axis_menu(m, ll, name, figobj, dir, idx):
            if len(ll) < 2:
                return m
            m.append((name, None, None))
            for i, x in enumerate(ll):
                def handler(evt, obj=figobj, name=x.name,
                            value=i, direction=dir):
                    canvas = evt.GetEventObject()
                    canvas.change_figobj_axes(figobj, value, direction)
                if i == idx:
                    m.append(('^'+x.name, handler, None))
                else:
                    m.append(('*'+x.name, handler, None))
            m.append(('!',  None, None))
            return m
        idx = self.get_axis_param_idx()
        m = []
        m = append_axis_menu(m, fig_axes._xaxis, '+X axis',
                             self, 'x', idx[0])
        m = append_axis_menu(m, fig_axes._yaxis, '+Y axis',
                             self, 'y', idx[1])
        m = append_axis_menu(m, fig_axes._caxis, '+C axis',
                             self, 'c', idx[3])

        return m

    def _export_shell(self, data, data_name, text):
        '''
        note: text is not used anymore
        '''
        import ifigure.server
        s = ifigure.server.Server()
        if s.info()[0]:
            print('Sending data to client()')
#           ifigure.server.Server().export_data(data)
            ifigure.server.Server().export_data((data, data_name,
                                                 text))
        text = '\n'
        for key in data:
            flag = self.write2shell(data[key], key)
            if flag is None:
                continue
            if flag:
                text = text + key + ' is updated\n'
            else:
                text = text + key + ' is created\n'
        dprint2(text)

        app = self.get_app()
        if app is not None:
            import ifigure.widgets.dialog as dialog
            ret = dialog.message(app, text, 'Export', 0)
            app.shell.SendShellEnterEvent()

    def onExport(self, event):
        canvas = event.GetEventObject()
        sel = [a() for a in canvas.selection]
        for a in self._artists:
            if a in sel:
                fig_val = self.get_export_val(a)
                self._export_shell(fig_val, 'fig_val', '')
        event.Skip()

    def export(self,  a=None):
        if a is None:
            sel = self._artists
        else:
            if not a in self._artists:
                return []
            sel = [a]

        return [self.get_export_val(a) for a in sel]

    def get_export_val(self, a):
        raise NotImplementedError(
            'get_export_val is missing in '+str(self.__class__))

    def damp_artists(self, fid, artists=None):
        name = self.name
        if artists is not None:
            pickle.dump({"num": len(artists),
                         "name": name}, fid)
        else:
            pickle.dump({"num": 0,
                         "name": name}, fid)

    def load_artists(self, fid, parent):
        # parent should be realized before....
        header = pickle.load(fid)
        self._artists = []
        # self._artists=[None]*header["num"]
        containers = parent.get_artist()
        self._container = containers[0]

        tmp = parent.get_childnames()
        if tmp.count(header["name"]) == 0:
            name = header["name"]
        else:
            name = self.get_next_name(self.get_namebase())

        parent.add_child(name, self)

#    this is import iscope....
#    def load(self, fin):
#        print  read_structure(fin)

    def add_figobj(self, name=None, obj=None):
        if obj is None:
            return
        if name is None:
            name = self.get_next_name("figobj")
        return self.add_child(name, obj)

    def del_figobj(self, figobj):
        figobj.destroy()
#        figobj.del_artist(delall=True)
#        self.del_child(figobj)

    def get_figobj(self, ifigobj):
        return self.get_child[ifigobj]

    def add_child(self, *args, **kargs):
        #
        #  figobj::add_child(name, obj, keep_zorder=bool)
        #  figobj::add_child(obj, keep_zorder=bool)
        #       1) set _zorder to the highest in page
        #       2) set axes bmp_update to false
        #
        if len(args) == 2:
            name = args[0]
            obj = args[1]
        elif len(args) == 1:
            obj = args[0]
            name = obj.name

        if 'keep_zorder' not in kargs:
            kargs['keep_zorder'] = False

        figpage = self.get_figpage()
        if not 'z_base' in kargs:
            if figpage is not None:
                z = max([obj2.getp('zorder')
                         for obj2 in figpage.walk_figobj()])
            else:
                z = 1
        else:
            z_base = kargs['z_base']
            z = z_base._z_base

        idx = super(FigObj, self).add_child(name, obj, **kargs)
#        self.get_child(idx=idx).set_bmp_update(False)
        obj.set_bmp_update(False)

        if kargs['keep_zorder']:
            return idx

#        for obj in self.get_child(idx).walk_figobj():
        for obj2 in obj.walk_figobj():
            obj2.setp('zorder', z+0.001)
            z = z+0.001

        if 'z_base' in kargs:
            z_base._z_base = z
        return idx
#
#   Attribute Setter and Getter
#     it "DOES NOT" copy data if all data
#     is taken at once

    def setp(self, *args):  # arg = name, var or var
        if len(args) == 2:
            self._attr[args[0]] = args[1]
        if len(args) == 1:
            self._attr = args[0]

    def getp(self, name=None):
        if name is None:
            return self._attr

        from ifigure.utils.cbook import isiterable_not_string

        if isiterable_not_string(name):
            try:
                return tuple([self._attr.get(n, None) for n in name])
            except:
                import sys
                import traceback
                print(("FigObj::getp error:", sys.exc_info()[0]))
                print(traceback.format_exc())
#              print "fig obj att not found "+name
                return [None]*len(name)
        else:
            try:
                return self._attr[name]
            except Exception:
                #          print "fig obj att not found "+name
                return None

    def delp(self, name=None):
        if name is None:
            return self._attr
        try:
            del self._attr[name]
        except Exception:
            #          print "fig obj att not found "+name
            return None

    def hasp(self, name):
        return name in self._attr


#
#  tree viewer menu
#


    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        #        if self._show_scriptmenu:
        base = super(FigObj, self).tree_viewer_menu()
#        else:
#           base=super(PyCode, self).tree_viewer_menu()
        return [('Show Attribute', self.onShowAtt, None),
                ('Reset Artist', self.onResetArtist, None),
                #                 ('+Evaluate',  None, None),
                #                 ('Replace',    self.onEvalReplace, None),
                #                 ('Append',     self.onEvalAppend, None),
                #                 ('Prepend',      self.onEvalFront, None),
                #                 ('FIFO',       self.onEvalFifo, None),
                #                 ('Config...',  self.onEvalConfig, None),
                #                 ('!',   None, None),
                ('---', None, None)]+base

    def eval_expressions(self, mode=''):
        if mode == '':
            mode = self._eval_mode
        names = self.dynamic_names()
        for name in names:
            v = self.eval(name, np=True)
            if mode == 'replace':
                self.setp(name, v)
            elif mode == 'append':
                v2 = np.concatenate((self.getp(name), v))
                self.setp(name, v2)
            elif mode == 'front':
                v2 = np.concatenate((v, self.getp(name)))
                self.setp(name, v2)
            elif mode == 'fifo':
                l = obj._eval_fifo_length
                print('do something')
                pass

    def onEvalReplace(self, evt=None):
        self.eval_expressions(mode='replace')
        self.del_artist(delall=True)
        self.generate_artist()
        ifigure.events.SendPVDrawRequest(self)

    def onEvalAppend(self, evt=None):
        self.eval_expressions(mode='append')
        self.del_artist(delall=True)
        self.generate_artist()
        ifigure.events.SendPVDrawRequest(self)

    def onEvalFront(self, evt=None):
        self.eval_expressions(mode='front')
        self.del_artist(delall=True)
        self.generate_artist()
        ifigure.events.SendPVDrawRequest(self)

    def onEvalFifo(self, evt=None):
        self.eval_expressions(mode='fifo')
        self.del_artist(delall=True)
        self.generate_artist()
        ifigure.events.SendPVDrawRequest(self)

    def onEvalConfig(self, evt=None):
        pass

    def get_root_figobj(self):
        '''
        find figobj which closest to root in the tree
        this obj is usually book (probably always?)
        '''
        def isFigObj(obj):
            from ifigure.mto.fig_obj import FigObj
            return isinstance(obj, FigObj)
        root = self.get_root_parent()
        for book in self.walk_tree_up(cond=isFigObj):
            pass
        return book

    def get_root_figgrp(self):
        '''
        return FigGrp close to p
        '''
        from ifigure.mto.fig_grp import FigGrp
        p = self
        in_group = False
        while p is not None:
            if isinstance(p, FigGrp):
                in_group = True
                break
            p = p.get_parent()

        if not in_group:
            return None
        while p is not None:
            if not isinstance(p.get_parent(), FigGrp):
                break
            p = p.get_parent()
        return p

    def onDelete(self, e=None):
        ifigure.events.SendPVDeleteFigobj(self)
        return

    def onResetArtist(self, e):
        root = self.get_root_parent()
        book = self.get_root_figobj()

        self.reset_artist()
        ifigure.events.SendPVDrawRequest(self)

    def onShowAtt(self, e):
        from matplotlib.artist import getp
        for key in self._attr:
            print((key, type(self._attr[key])))
        for a in self._artists:
            print(getp(a))

    def onForceUpdate(self, evt):
        print('calling forced update.... check if it is necessary')
        id = evt.GetEventObject()
        app = id.GetTopLevelParent()
        app.deffered_force_layout()

    def onSuppress(self, evt=None):
        self.set_suppress(True)
#        self.realize()
        ifigure.events.SendPVDrawRequest(self)
#        app.draw()
        if evt is not None:
            evt.Skip()

    def onUnSuppress(self, evt=None):
        #        id=evt.GetEventObject()
        #        app=id.GetTopLevelParent()
        self.set_suppress(False)
#        self.realize()
        ifigure.events.SendPVDrawRequest(self)
#        app.draw()
        if evt is not None:
            evt.Skip()

    def set_suppress(self, val):
        super(FigObj, self).set_suppress(val)
        self.realize()
        self.set_bmp_update(False)
###
# plot mode picking/dragging
# the behavior of this mode should be implemented
# in subclasses
###
# Hit test

    def picker(self, artist, evt):
        print('picker is obsolete and should not be called')
        return False, {}
# dragstart : called when drag starts

    def dragstart(self, a, evt):
        redraw = 0  # request to draw canvas 0|1|??
        return redraw
# drag : to show transient usr feedback during drag

    def drag(self, a, evt):
        redraw = 0  # request to draw canvas 0|1|??
        return redraw
# dragdone : finish-up dragging

    def dragdone(self, a, evt):
        redraw = 0  # request to draw canvas 0|1|??
        return redraw

# drag_get_hl (return hl artist)
    def drag_get_hl(self, a):
        return []

    def drag_rm_hl(self, a):
        pass


# Hit test (annotation mode)

    @property
    def ispickable_a(self):
        return True

    def picker_a(self, a, evt):
        hit, extra, type, loc = self.picker_a0(a, evt)
        if hit:
            self._picker_a_type = type
            self._picker_a_loc = loc
        return hit, extra

    def get_artist_extent2(self, a):
        return self.get_artist_extent(a)

    def get_artist_extent(self, a):
        '''
        retrun the extent of artist in device 
        coordinate
        '''
        try:
            #box = a.get_window_extent(a.figure._cachedRenderer)
            box = a.get_window_extent(a.figure.canvas.get_renderer())
            return box.xmin, box.xmax, box.ymin, box.ymax
        except:
            print(('error in get_artist_extent for ', a))
            print(traceback.format_exc())
            return [None]*4
        return [None]*4

    def picker_a0(self, a, evt):
        '''
        generic annotation picker. it check hit to
        rect returned by get_artist_extent

        return value : hit, {}, type, loc
           type : 'area'   aloc =  3: hit in both xy, 1: x hit , 2 y hit
                  'edge'   loc = {0,1,2,3} 
                  'point'  loc = {0,1,2,3} 

        '''
        def checker(a0, b0, b1):
            if (a0 > b0 and a0 < b1):
                return True
            return False
        self._st_p = evt.x, evt.y
        x0d, x1d, y0d, y1d = self.get_artist_extent(a)
        if x0d is None:
            return False, {}, -1, -1
        hit_d = 5

        aloc = 0
        if checker(evt.x, x0d+hit_d, x1d-hit_d):
            aloc = aloc + 1
        if checker(evt.y, y0d+hit_d, y1d-hit_d):
            aloc = aloc + 2
        eloc = 0
        if checker(evt.x, x0d-hit_d, x0d+hit_d):
            eloc = eloc + 1
        if checker(evt.x, x1d-hit_d, x1d+hit_d):
            eloc = eloc + 2
        if checker(evt.y, y0d-hit_d, y0d+hit_d):
            eloc = eloc + 4
        if checker(evt.y, y1d-hit_d, y1d+hit_d):
            eloc = eloc + 8

        if aloc == 0 and eloc == 0:
            type = None
            return False, {}, type, 0
#        print aloc, eloc
        if aloc == 3:
            type = 'area'
            return True, {}, type, aloc
        if ((eloc & 1 != 0 or eloc & 2 != 0) and
                aloc == 2):
            type = 'edge'
            return True, {}, type, eloc
        if ((eloc & 4 != 0 or eloc & 8 != 0) and
                aloc == 1):
            type = 'edge'
            return True, {}, type, eloc
        if (eloc != 1 and eloc != 2 and eloc != 4 and eloc != 8
                and aloc == 0):
            type = 'point'
            return True, {}, type, eloc

        return False, {}, None, 0

# dragstart : called when drag starts
    def dragstart_a(self, a, evt, mode=1):
        '''
        fig_obj::dragstart_a(self, a, evt)
        fig_obj::drag_a(self, a, evt, shift = None)
        fig_obj::dragdone_a(self, a, evt, shift = None)
        these three methods provide generic drag handling 
        in annnotation mode. it calculate new window_extent
        and save it in propoerty. derived classes
        can used it to provide "live-update" of object 
        while drag is happening.
        '''
        redraw = 0
        if mode == 1:
            self._st_extent = self.get_artist_extent(a)

#        self.highlight_artist(False, artist=[a])

        x = [self._st_extent[0],
             self._st_extent[0],
             self._st_extent[1],
             self._st_extent[1],
             self._st_extent[0]]
        y = [self._st_extent[2],
             self._st_extent[3],
             self._st_extent[3],
             self._st_extent[2],
             self._st_extent[2]]
        self.drag_a_add_hl(a, x, y)
#        if self._drag_hl is None:

        self._drag_a_data = None
        return redraw

    # drag : to show transient usr feedback during drag
    def drag_a(self, a, evt, shift=None, scale=None):
        '''
        fig_obj::dragstart_a(self, a, evt)
        fig_obj::drag_a(self, a, evt, shift = None)
        fig_obj::dragdone_a(self, a, evt, shift = None)
        these three methods provide generic drag handling 
        in annnotation mode. it calculate new window_extent
        and save it in propoerty. derived classes
        can used it to provide "live-update" of object 
        while drag is happening.
        '''
        if shift is None:
            shift = evt.guiEvent.ShiftDown()

        loc = self._picker_a_loc
        rec = [x for x in self._st_extent]

        if scale is None:
            if self._picker_a_type == 'area':
                dx = evt.x - self._st_p[0]
                dy = evt.y - self._st_p[1]
                if shift:
                    if abs(dx) < abs(dy):
                        dx = 0
                    else:
                        dy = 0
                rec[0] = rec[0]+dx
                rec[1] = rec[1]+dx
                rec[2] = rec[2]+dy
                rec[3] = rec[3]+dy

            else:
                if ((loc & 1) != 0):
                    rec[0] = evt.x
                if ((loc & 2) != 0):
                    rec[1] = evt.x
                if ((loc & 4) != 0):
                    rec[2] = evt.y
                if ((loc & 8) != 0):
                    rec[3] = evt.y

            if shift and self._picker_a_type != 'area':
                if self._st_extent[1]-self._st_extent[0] != 0:
                    d = (float(self._st_extent[3]-self._st_extent[2]) /
                         float(self._st_extent[1]-self._st_extent[0]))
                else:
                    d = 1
                dy = float(rec[1]-rec[0])*d
                if ((loc & 4) != 0):
                    rec[2] = int(rec[3]-dy)
                if ((loc & 8) != 0):
                    rec[3] = int(rec[2]+dy)

            self._drag_rec = rec
            scale = geom.calc_scale(rec, self._st_extent)
        else:
            rec = geom.scale_rect(rec, scale)
            self._drag_rec = rec

        x = [rec[0], rec[0], rec[1], rec[1], rec[0]]
        y = [rec[2], rec[3], rec[3], rec[2], rec[2]]
        self.drag_a_add_hl(a, x, y)
#        self._drag_hl.set_xdata(x)
#        self._drag_hl.set_ydata(y)

        redraw = 0  # request to draw canvas 0|1|??

        self._drag_a_data = redraw, scale
        return redraw, scale

    def drag_a_get_hl(self, a):
        if self._drag_hl is None:
            return []
        return [self._drag_hl]

    def drag_a_add_hl(self, a, x, y):
        if self._drag_hl is not None:
            self.drag_a_rm_hl(a)
        self._drag_hl = Line2D(x, y, figure=a.figure,
                               marker='s',
                               linestyle='-',
                               markerfacecolor='None')
        a.figure.lines.extend([self._drag_hl])

    def drag_a_rm_hl(self, a):
        if self._drag_hl is not None:
            a.figure.lines.remove(self._drag_hl)
            self._drag_hl = None

    def dragdone_a_clean(self, a):
        pass

# dragdone : finish-up dragging
    def dragdone_a(self, a, evt, shift=None, scale=None):
        '''
        fig_obj::dragstart_a(self, a, evt)
        fig_obj::drag_a(self, a, evt, shift = None)
        fig_obj::dragdone_a(self, a, evt, shift = None)
        these three methods provide generic drag handling 
        in annnotation mode. it calculate new window_extent
        and save it in propoerty. derived classes
        can used it to provide "live-update" of object 
        while drag is happening.
        '''
        if self._drag_a_data is not None:
            redraw, scale = self._drag_a_data
        else:
            # fall back...just in case
            redraw, scale = self.drag_a(a, evt, shift=shift, scale=scale)
        self.drag_a_rm_hl(a)

        return 0, scale

# scale_artist : scale artist by using sclae...
    def scale_artist(self, scale, action=True):
        pass

###   group/ungroup (ungroup is defined in FigGrp)
    def group(self, member):
        '''
        group figgrp object
        member : list of treedict to be grouped
                 (self is automatically added to 
                  member)
        return a FigGrp containes member. 
        if self has a parent. FigGrp will be
        added to the parent
        '''
        from ifigure.mto.fig_grp import FigGrp
        gp = FigGrp()
        print(member)
        p = self.get_parent()
        if p is not None:
            p.add_child(p.get_next_name(gp.get_namebase()),
                        gp, keep_zorder=True)

        if not self in member:
            member = [self]+member
        for obj in member:
            obj.move(gp, keep_zorder=True)
        return gp

# save/load
#    def save_data(self, fid=None):
#        val={"num_artists":len(self._artists),
#             "format": 2}
#        pickle.dump(val, fid)
#        for a in self._artists:
#           val=self.get_artist_property(a)
#           pickle.dump(val, fid)

#        ps = self.attr_in_file()
#        val = {}
#        for name in ps:
#           val[name]=self.getp(name)

#        secret="loaded_property"
#        if self.hasp(secret):
#           val[secret]=self.getp(secret)
#        pickle.dump(val, fid)

    def store_loaded_property(self):
        if self.hasp("loaded_property"):
            return False
        if (not self.isempty() and
                not self.hasp("loaded_property")):
            ap = [self.get_artist_property(a) for a in self._artists]
            self.setp("loaded_property", ap)
        if (self.isempty() and
                not self.hasp("loaded_property")):
            self.setp("loaded_property", [])
        return True

    def save_data2(self, data=None):
        if data is None:
            data = {}
        #ap = [self.get_artist_property(a) for a in self._artists]
        # artist_property is stored always as "loaded_property"
        # ap is left for backword compatibility
        ap = []
        is_new_stored_prop = self.store_loaded_property()
        ###
#        ps = properties_in_file_1(self)
        ps = self.attr_in_file()
        val = {}
        for name in ps:
            val[name] = self.getp(name)

        # this section is always True
        if self.hasp("loaded_property"):
            val["loaded_property"] = self.getp("loaded_property")

        param = {"_container_idx": self._container_idx}
        data['FigObj'] = (1, ap, val, param)

        # this is skipped, since superclass is treedict
        data = super(FigObj, self).save_data2(data)
        if is_new_stored_prop:
            self.delp('loaded_property')
        return data

    def handle_loaded_figobj_data(self, data, clsname):
        if clsname in data:
            val = data[clsname][1]
            for key in val:
                self.setp(key, val[key])
            if len(data[clsname]) > 2:
                param = data[clsname][2]
                for key in param:
                    if hasattr(self, key):
                        setattr(self, key, param[key])

    def load_data(self, fid=None):
        val = pickle.load(fid)
        loaded_prop = []
        for i in range(0, val["num_artists"]):
            prop = pickle.load(fid)
            loaded_prop.append(prop)
        if len(loaded_prop) != 0:
            self.setp("loaded_property", loaded_prop)
        else:
            pass
#           print "no loaded property....artist was not realized when saved?"
#        print(loaded_prop)
        if "format" in val:
            if val["format"] == 2:
                attr = pickle.load(fid)
                for k in attr:
                    self.setp(k, attr[k])
        if not self.hasvar("kywds"):
            self.setvar("kywds", {})

    def load_data2(self, data):
        d = data['FigObj']
        loaded_prop = d[1]
        attr = d[2]
        if len(d) > 3:
            param = d[3]
            for key in param:
                if hasattr(self, key):
                    setattr(self, key, param[key])
        if len(loaded_prop) != 0:
            self.setp("loaded_property", loaded_prop)
        else:
            pass
#           print "no loaded property....artist was not realized when saved?"

        for k in attr:
            self.setp(k, attr[k])
        if not self.hasvar("kywds"):
            self.setvar("kywds", {})
        super(FigObj, self).load_data2(data)


#    def init_after_load(self, olist, nlist):
#        from ifigure.mto.axis_user import XUser, YUser, ZUser, CUser
#        if isinstance(self, XUser): self.get_xaxisparam()
#        if isinstance(self, YUser): self.get_yaxisparam()
#        if isinstance(self, ZUser): self.get_zaxisparam()
#        if isinstance(self, CUser): self.get_caxisparam()

    def get_artist_property(self, a):
        #       plist = properties_in_file_0(self)
        plist = self.property_in_file()
        vals = {}
        for p in plist:
            p0 = p
            if isMPL2 and p == 'axis_bgcolor':
                p = 'facecolor'
            if hasattr(a, 'get_'+p):
                if callable(getattr(a, 'get_'+p)):
                    vals[p0] = (getattr(a, 'get_'+p))()
                else:
                    vals[p0] = getattr(a, 'get_'+p)
            elif hasattr(a, p):
                vals[p0] = getattr(a, p)
        return vals

    def set_artist_property(self, a, vals):
        for key in vals:
            #          if isMPL2 and key == 'axis_bgcolor': key = 'facecolor'
            if hasattr(a, 'set_'+key):
                if callable(getattr(a, 'set_'+key)):
                    (getattr(a, 'set_'+key))(vals[key])
                else:
                    setattr(a, 'set_'+key, vals[key])
            elif hasattr(a,  key):
                setattr(a, key, vals[key])

    def switch_scale(self, level):
        pass

#
#   set artist property
#
    def mpl_set(self, name, *value, **kargs):
        if "ia" in kargs:
            ia = kargs["ia"]
        else:
            ia = 0
        ax = self.get_figaxes()
        if ax is not None:
            ax.set_bmp_update(False)
        self.set_client_update_artist_request()
        if len(value) == 1:
            return self._artists[ia].set(**{name: value[0]})
        else:
            return self._artists[ia].set(**{name: value})

    def mpl_get(self, name, ia=0):
        m = getattr(self._artists[ia], 'get_'+name)
        return m()


#
#   range setting
#

    def get_xrange(self, xrange=[None, None], scale='linear'):
        return xrange

    def get_yrange(self, yrange=[None, None],
                   xrange=[None, None], scale='linear'):
        return yrange

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], sclae='linear'):
        return crange

    def get_zrange(self, zrange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None], scale='linear'):
        return zrange

    def _update_range(self, range, de):
        masked = np.ma.core.MaskedConstant
        if not isinstance(de[0], masked):
            if range[0] is None:
                range[0] = de[0]
            else:
                if de[0] is not None:
                    range[0] = min([range[0], de[0]])
        if not isinstance(de[1], masked):
            if range[1] is None:
                range[1] = de[1]
            else:
                if de[1] is not None:
                    range[1] = max([range[1], de[1]])
        return range
#
#
#

    def convert_to_tex_style_text(self, mode=True):
        pass
#
#   axes change
#

    def handle_axes_change(self, evt):
        '''
        some objects requires to change artist 
        if scale/range changed
        '''
        pass
#
#  axes, transform...
#

    def set_axesref(self, figaxes_list):
        self._fig_axes = []
        for figaxes in figaxes_list:
            if figaxes is not None:
                self._fig_axes.append(weakref.ref(figaxes, self.lost_figaxes))
            else:
                self._fig_axes.append(cbook.WeakNone())

    def get_axesref(self):
        return self._fig_axes

    def lost_figaxes(self, ref):
        print('referred axes is deleted')

    def convert_transform_rect(self, t1, t2, rect):
        n1 = np.array([rect[0], rect[1]]).reshape(1, 2)
        n2 = np.array([rect[0]+rect[2], rect[1]+rect[3]]).reshape(1, 2)
        if t1 is not None:
            n1 = t1.transform(n1)
            n2 = t1.transform(n2)
        if t2 is not None:
            n1 = t2.inverted().transform(n1)
            n2 = t2.inverted().transform(n2)

        w = n2[0][0]-n1[0][0]
        h = n2[0][1]-n1[0][1]
        rect = [n1[0][0], n1[0][1], w, h]

        return rect

    def convert_transform_point(self, t1, t2, p):
        '''
        it dose 
        1) transfomr to device using t1
        2) transform to new coord by t2.inverted()
        '''
        n1 = np.array([p[0], p[1]]).reshape(1, 2)
        if t1 is not None:
            n1 = t1.transform(n1)
        if t2 is not None:
            n1 = t2.inverted().transform(n1)
        return n1[0][0], n1[0][1]

    def coordsname2transform(self, fig_page, name, fig_axes=None):
        #name = name0[0]
        if fig_axes is None:
            return fig_page._artists[0].transFigure
        if name == 'data':
            return fig_axes._artists[0].transData
        elif name == 'axes':
            return fig_axes._artists[0].transAxes
        elif name == 'figure':
            return fig_page._artists[0].transFigure

#
# change rotation center in 3D plot
#
    def onSetRotCenter(self, evt):
        canvas = evt.GetEventObject()

        xx = 0
        yy = 0
        zz = 0
        ll = 0
        for x in canvas.selection:
            glartist = x()
            x, y, z = glartist.get_glverts()
            l = len(x)
            ll += l
            xx += np.mean(x)*l
            yy += np.mean(y)*l
            zz += np.mean(z)*l
            axes = glartist.axes

        xx /= l
        yy /= l
        zz /= l

        cc = np.array((xx, yy, zz))
        axes._gl_rot_center = cc
        axes._gl_use_rot_center = True

        evt.Skip()
